from .._grammar import select
from ._char_range import char_range


def char_set(def_string: str):
    parts = []
    pos = 0
    while pos < len(def_string):
        if pos + 2 < len(def_string) and def_string[pos + 1] == "-":
            parts.append(char_range(def_string[pos], def_string[pos + 2]))
            pos += 3
        elif pos + 1 < len(def_string) and def_string[pos] == "\\":
            parts.append(def_string[pos + 1])
            pos += 2
        else:
            parts.append(def_string[pos])
            pos += 1
    return select(parts)



import tiktoken

from ._model import Chat, Instruct
from ._grammarless import GrammarlessTokenizer, GrammarlessEngine, Grammarless


class LiteLLMEngine(GrammarlessEngine):
    def __init__(
        self,
        model,
        tokenizer,
        timeout,
        compute_log_probs,
        max_streaming_tokens,
        **kwargs,
    ):
        try:
            import litellm
        except ModuleNotFoundError:
            raise Exception(
                "Please install the litellm package version >= 1.7 using `pip install litellm -U` in order to use guidance.models.LiteLLM!"
            )

        self.litellm = litellm

        # self.client = openai_package.OpenAI(api_key=api_key, organization=organization, base_url=base_url)
        self.model_name = model

        # we pretend it tokenizes like gpt2 if tiktoken does not know about it... TODO: make this better
        if tokenizer is None:
            try:
                tokenizer = tiktoken.encoding_for_model(model)
            except:
                tokenizer = tiktoken.get_encoding("gpt2")

        super().__init__(
            tokenizer,
            max_streaming_tokens=max_streaming_tokens,
            timeout=timeout,
            compute_log_probs=compute_log_probs,
        )


class LiteLLM(Grammarless):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        timeout=0.5,
        max_streaming_tokens=1000,
        compute_log_probs=False,
    ):
        """Build a new LiteLLM model object that represents a model in a given state."""

        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is LiteLLM:
            raise Exception(
                "The LightLLM class is not meant to be used directly! Please use LiteLLMChat, LiteLLMInstruct, or LiteLLMCompletion depending on the model you are using."
            )

        # this allows us to use a single constructor for all our subclasses
        engine_map = {
            LiteLLMCompletion: LiteLLMCompletionEngine,
            LiteLLMInstruct: LiteLLMInstructEngine,
            LiteLLMChat: LiteLLMChatEngine,
        }

        for k in engine_map:
            if issubclass(self.__class__, k):
                super().__init__(
                    engine_map[k](
                        model,
                        tokenizer,
                        timeout,
                        compute_log_probs,
                        max_streaming_tokens,
                    ),
                    echo=echo,
                )


class LiteLLMCompletion(LiteLLM):
    pass


class LiteLLMCompletionEngine(LiteLLMEngine):

    def _generator(self, prompt, temperature):

        # update our shared data state
        self._reset_shared_data(prompt, temperature)

        try:
            generator = self.litellm.completion(
                model=self.model_name,
                messages=[
                    {"content": prompt.decode("utf8"), "role": "system"}
                ],  # note that role=system is just ignored by litellm but used by them to match chat syntax
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1,
                temperature=temperature,
                stream=True,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for part in generator:
            chunk = part.choices[0].delta.content or ""
            yield chunk.encode("utf8")


class LiteLLMInstruct(LiteLLM, Instruct):
    def get_role_start(self, name):
        return ""

    def get_role_end(self, name):
        if name == "instruction":
            return "<|endofprompt|>"
        else:
            raise Exception(
                f"The LiteLLMInstruct model does not know about the {name} role type!"
            )


class LiteLLMInstructEngine(LiteLLMEngine):

    def _generator(self, prompt, temperature):
        # start the new stream
        prompt_end = prompt.find(b"<|endofprompt|>")
        if prompt_end >= 0:
            stripped_prompt = prompt[:prompt_end]
        else:
            raise Exception(
                "This model cannot handle prompts that don't match the instruct format!"
            )

        # make sure you don't try and instruct the same model twice
        if b"<|endofprompt|>" in prompt[prompt_end + len(b"<|endofprompt|>") :]:
            raise Exception(
                "This model has been given two separate instruct blocks, but this is not allowed!"
            )

        # update our shared data state
        self._reset_shared_data(stripped_prompt + b"<|endofprompt|>", temperature)

        try:
            generator = self.litellm.completion(
                model=self.model_name,
                messages=[
                    {"content": self._data.decode("utf8"), "role": "system"}
                ],  # note that role=system is just ignored by litellm but used by them to match chat syntax
                prompt=self._data.decode("utf8"),
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1,
                temperature=temperature,
                stream=True,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for part in generator:
            chunk = part.choices[0].delta.content or ""
            yield chunk.encode("utf8")


class LiteLLMChat(LiteLLM, Chat):
    pass


class LiteLLMChatEngine(LiteLLMEngine):
    def _generator(self, prompt, temperature):

        # find the system text
        pos = 0
        role_end = b"<|im_end|>"

        # find the user/assistant pairs
        messages = []
        found = True
        while found:

            # find the user text
            found = False
            for role_name, start_bytes in (
                ("system", b"<|im_start|>system\n"),
                ("user", b"<|im_start|>user\n"),
                ("assistant", b"<|im_start|>assistant\n"),
            ):
                if prompt[pos:].startswith(start_bytes):
                    pos += len(start_bytes)
                    end_pos = prompt[pos:].find(role_end)
                    if end_pos < 0:
                        assert (
                            role_name == "assistant"
                        ), "Bad chat format! Last role before gen needs to be assistant!"
                        break
                    btext = prompt[pos : pos + end_pos]
                    pos += end_pos + len(role_end)
                    messages.append(
                        {"role": role_name, "content": btext.decode("utf8")}
                    )
                    found = True
                    break

        # update our shared data state
        self._reset_shared_data(prompt[:pos], temperature)

        try:
            generator = self.litellm.completion(
                model=self.model_name,
                messages=messages,
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1,
                temperature=temperature,
                stream=True,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for part in generator:
            chunk = part.choices[0].delta.content or ""
            yield chunk.encode("utf8")



import re
import types

from typing import Any, Dict, List, TYPE_CHECKING, TypeVar, Union

from . import _serialization_pb2
from . import _parser

_T = TypeVar("_T")

# to support the embedding of guidance functions inside Python f-strings we use tags with these delimiters
tag_start = "{{G|"  # start of a call tag
tag_end = "|G}}"  # end of a call tag
_call_pool: Dict[str, "Function"] = {}  # the functions associated with the call tags
_tag_pattern = re.compile(
    re.escape(tag_start) + r"([^\|]+)" + re.escape(tag_end)
)  # the pattern for matching call tags

class StatefulException(Exception):
    """This is raised when we try and use the state of a grammar object like it was a live model.

    Note that eventually it would be nice to support stateful parser/grammar constructs directly, but
    such "parser combinators" cannot be run effciently in Python. So we use a traditional parser and
    grammar separation (hence the need for this exception)."""

    pass


class Function:
    """This is the abstract class representing all guidance functions.

    There are two main subclasses: GrammarFunction and RawFunction. GrammarFunctions
    represent guidance grammars that can be serialized and sent across the wire, while
    RawFunctions represent unconstrained native Python functions.
    """

    def __init__(self, name, value=None) -> None:
        self.name = name
        self.value = value

    def __str__(self):
        """Creates a string tag that can be used to retrieve this object."""

        # save the call in our call pool, ready to be run when it is attached to an LM object
        str_id = str(id(self))
        if str_id not in _call_pool:
            _call_pool[str_id] = self

        # return a string representation of this call so it can be combined with other strings/calls
        return tag_start + str_id + tag_end

    def serialize(self):
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, serialized_grammar):
        raise NotImplementedError()


class RawFunction(Function):
    __slots__ = ("f", "args", "kwargs")

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self, model):
        return self.f(model, *self.args, **self.kwargs)

    def __add__(self, other):

        # if we are joining with a string we use the string representation for ourselves
        if isinstance(other, str):
            return str(self) + other

        def __add__(model):
            model = self(model)
            if model is None:
                raise Exception(
                    f"The guidance function `{self.f.__name__}` did not return a model object! You need to return an updated model object at the end of your guidance function."
                )
            if isinstance(other, GrammarFunction):
                return model + other
            else:
                return other(model)

        return RawFunction(__add__, [], {})

    def __radd__(self, other):

        # if we are joining with a string we use the string representation for ourselves
        if isinstance(other, str):
            return other + str(self)

        def __radd__(model):
            if isinstance(other, GrammarFunction):
                model += other
            else:
                model = other(model)
            return self(model)

        return RawFunction(__radd__, [], {})


class Match:
    def __init__(self, captures, log_probs, partial):
        self.captures = captures
        self.log_probs = log_probs
        self.partial = partial

    def __getitem__(self, key):
        return self.captures[key]

    def __len__(self):
        return len(self.captures)

    def __bool__(self):
        return True

    def __str__(self):
        return str(self.captures)

    def __repr__(self):
        return (
            "<guidance.Match object; captures="
            + str(self.captures)
            + "; partial="
            + str(self.partial)
            + ">"
        )


class GrammarFunction(Function):
    num_used_names = 0

    def __add__(self, value):

        # see if we have a string with calls or a simple string
        if isinstance(value, str) or isinstance(value, bytes):
            if isinstance(value, str) and re.search(_tag_pattern, value):
                return str(self) + value
            else:
                value = string(value)

        # see if we can keep building a stateless grammar
        if isinstance(value, GrammarFunction):
            return Join([self, value])

        # otherwise we let the stateful object handle things
        else:
            return value.__radd__(self)

    def __radd__(self, value):

        # see if we have a string with calls or a simple string
        if isinstance(value, str) or isinstance(value, bytes):
            if isinstance(value, str) and re.search(_tag_pattern, value):
                return value + str(self)
            else:
                value = string(value)

        # see if we can keep building a stateless grammar
        if isinstance(value, GrammarFunction):
            return Join([value, self])

        # otherwise we let the stateful object handle things
        else:
            return value.__add__(self)

    def __getitem__(self, value):
        raise StatefulException("GrammarFunctions can't access state!")

    def match(
        self,
        byte_string: Union[str, bytes],
        allow_partial: bool = False,
        raise_exceptions: bool = False,
    ) -> Union[Match, None]:
        if isinstance(byte_string, str):
            byte_string = byte_string.encode()
        parser = _parser.EarleyCommitParser(self)

        for i in range(len(byte_string)):
            try:
                parser.consume_byte(byte_string[i : i + 1])
            except _parser.ParserException:
                if raise_exceptions:
                    raise
                else:
                    return None

        if not allow_partial and not parser.matched():
            return None
        else:
            return Match(*parser.get_captures(), partial=not parser.matched())  # type: ignore[misc]

    @staticmethod
    def _new_name():
        num_used = GrammarFunction.num_used_names

        a_ord = ord("a")

        # name the name in base 26 letter notation
        name = chr(a_ord + num_used % 26)
        if num_used >= 26:
            name = chr(a_ord + (num_used % 676) // 26) + name
            if num_used >= 676:
                name = chr(a_ord + (num_used % 17576) // 676) + name
                if num_used >= 17576:
                    name = chr(a_ord + (num_used % 456976) // 17576) + name

        GrammarFunction.num_used_names += 1

        return name

    def gbnf_string(self):
        used_names = set()
        names = {}
        lines = []
        root_name = self._rec_gbnf_string(lines, used_names, names)
        lines.append("root ::= " + root_name)
        return "\n".join(lines)

    def serialize(self):
        g = _serialization_pb2.Grammar()
        index_map = {}
        nodes = {}
        self._rec_create_index_map(index_map)  # gives all the nodes an index
        self._rec_serialize(index_map, nodes)  # nodes is filled in (as is index_map)
        g.nodes.extend(list(nodes.values()))
        return g.SerializeToString()

    def _rec_create_index_map(self, index_map):
        if self not in index_map:
            index_map[self] = len(index_map)
            if hasattr(self, "values"):
                for value in self.values:
                    value._rec_create_index_map(index_map)

    def _rec_serialize(self, index_map, nodes):
        if self not in nodes:
            v = self._to_proto(index_map)
            node = _serialization_pb2.GrammarFunction()
            if isinstance(self, Byte):
                node.byte.CopyFrom(v)
            elif isinstance(self, ByteRange):
                node.byte_range.CopyFrom(v)
            elif isinstance(self, Select):
                node.select.CopyFrom(v)
            elif isinstance(self, Join):
                node.join.CopyFrom(v)
            elif isinstance(self, ModelVariable):
                node.model_variable.CopyFrom(v)
            else:
                raise Exception("Unknown node type")
            nodes[self] = node
            if hasattr(self, "values"):
                for value in self.values:
                    value._rec_serialize(index_map, nodes)

    @classmethod
    def deserialize(cls, serialized_grammar):
        g = _serialization_pb2.Grammar()
        g.ParseFromString(serialized_grammar)

        # create the list of objects
        values = []
        for node in g.nodes:
            if node.HasField("byte"):
                node = Byte._from_proto(node.byte)
            elif node.HasField("byte_range"):
                node = ByteRange._from_proto(node.byte_range)
            elif node.HasField("select"):
                node = Select._from_proto(node.select)
            elif node.HasField("join"):
                node = Join._from_proto(node.join)
            elif node.HasField("model_variable"):
                node = ModelVariable._from_proto(node.model_variable)
            else:
                raise Exception("Unknown node type")
            values.append(node)

        # fill in the values pointers now that we have the full list of objects
        for v in values:
            if hasattr(v, "values"):
                for i, index in enumerate(v.values):
                    v.values[i] = values[index]

        return values[0]  # the first element in the root node of the grammar


class Terminal(GrammarFunction):
    def match_byte(self, byte):
        pass  # abstract

    @property
    def max_tokens(self):
        return 1000000000000


class Byte(Terminal):
    __slots__ = ("byte", "hidden", "commit_point", "capture_name", "temperature")

    def __init__(self, byte):
        assert isinstance(byte, bytes)
        assert len(byte) == 1
        self.byte = byte
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.temperature = -1

    @property
    def name(self):
        return str(self.byte)

    def __hash__(self):
        return self.byte[0]

    def __eq__(self, other):
        return isinstance(other, Byte) and self.byte[0] == other.byte[0]

    def __repr__(self) -> str:
        return str(self.byte)

    def __len__(self):
        return 1

    def match_byte(self, byte):
        return byte == self.byte

    @property
    def nullable(self):
        return False

    def _to_proto(self, index_map):
        data = _serialization_pb2.Byte()
        data.byte = self.byte
        data.hidden = self.hidden
        data.commit_point = self.commit_point
        data.capture_name = "" if self.capture_name is None else self.capture_name
        data.temperature = self.temperature
        return data

    @staticmethod
    def _from_proto(data):
        out = Byte(data.byte)
        out.hidden = data.hidden
        out.commit_point = data.commit_point
        out.capture_name = None if data.capture_name == "" else data.capture_name
        out.temperature = data.temperature
        return out


class ByteRange(Terminal):
    __slots__ = ("byte_range", "hidden", "commit_point", "capture_name", "temperature")

    def __init__(self, byte_range):
        assert isinstance(byte_range, bytes)
        assert len(byte_range) == 2
        self.byte_range = byte_range
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.temperature = -1  # -1 means not set

    def match_byte(self, byte):
        return self.byte_range[0] <= byte[0] <= self.byte_range[1]

    @property
    def name(self):
        return str(self.byte_range)

    @name.setter
    def name(self, value):
        pass  # we ignore name changes

    @property
    def nullable(self):
        return False

    def __hash__(self):
        return self.byte_range[0] + 256 * self.byte_range[1]

    def __eq__(self, other):
        return (
            isinstance(other, ByteRange)
            and self.byte_range[0] == other.byte_range[0]
            and self.byte_range[1] == other.byte_range[1]
        )

    def __repr__(self) -> str:
        return str(self.byte_range)

    def __len__(self):
        return 1

    def _to_proto(self, index_map):
        data = _serialization_pb2.ByteRange()
        data.byte_range = self.byte_range
        data.hidden = self.hidden
        data.commit_point = self.commit_point
        data.capture_name = "" if self.capture_name is None else self.capture_name
        data.temperature = self.temperature
        return data

    @staticmethod
    def _from_proto(data):
        out = ByteRange(data.byte_range)
        out.hidden = data.hidden
        out.commit_point = data.commit_point
        out.capture_name = None if data.capture_name == "" else data.capture_name
        out.temperature = data.temperature
        return out


class Null(Terminal):
    __slots__ = ("name", "hidden", "commit_point", "capture_name")

    nullable = True

    def __init__(self):
        self.name = None
        self.hidden = False
        self.commit_point = False
        self.capture_name = None

    def __add__(self, other):
        # see if we have a string with calls or a simple string
        if isinstance(other, bytes):
            return string(other)
        elif isinstance(other, str):
            return str_to_grammar(other)

        # otherwise we return unchanged
        else:
            return other

    def __radd__(self, other):
        return self.__add__(
            other
        )  # left vs right makes no difference since we are null


class ModelVariable(GrammarFunction):
    """This represents a variable that will be read from the model object when this grammar is executed.

    Note that the name is the name of the attribute on the model object this node
    will get replaced with.
    """

    __slots__ = ("name", "hidden", "commit_point", "capture_name")

    def __init__(self, name):
        self.name = name
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.nullable = False

    def _to_proto(self, index_map):
        data = _serialization_pb2.ModelVariable()
        data.hidden = self.hidden
        data.name = self.name
        data.commit_point = self.commit_point
        data.capture_name = "" if self.capture_name is None else self.capture_name
        return data

    @staticmethod
    def _from_proto(data):
        out = ModelVariable(data.name)
        out.hidden = data.hidden
        out.commit_point = data.commit_point
        out.capture_name = None if data.capture_name == "" else data.capture_name
        return out


def replace_grammar_node(grammar, target, replacement):
    # Use a stack to keep track of the nodes to be visited
    stack = [grammar]
    visited_set = set()  # use set for O(1) lookups

    while stack:
        current = stack.pop()

        # Check if we have already visited this node
        if current in visited_set:
            continue
        visited_set.add(current)

        # We are done with this node if it's a terminal
        if isinstance(current, (Terminal, ModelVariable, Placeholder)):
            continue

        # Iterate through the node's values and replace target with replacement
        for i, value in enumerate(current.values):
            if value == target:
                current.values[i] = replacement
            else:
                stack.append(value)


# def replace_grammar_node(grammar, target, replacement, visited_set={}):

#     # see if we have already visited this node
#     if grammar in visited_set:
#         return
#     else:
#         visited_set[grammar] = True

#     # we are done if this is a terminal
#     if isinstance(grammar, (Terminal, ModelVariable)):
#         return

#     # replace all matching sub-nodes
#     for i,value in enumerate(grammar.values):
#         if value == target:
#             grammar.values[i] = replacement
#         else:
#             replace_grammar_node(value, target, replacement, visited_set)


def replace_model_variables(grammar, model, allowed_vars=None):
    """Replace all the ModelVariable nodes with their values in an iterative manner."""
    visited_set = set()
    stack = [
        (grammar, None, None)
    ]  # Stack stores tuples of (node, parent_node, child_index)
    replacements = []

    while stack:
        current, parent, child_index = stack.pop()

        # This node is being visited for the first time
        if current not in visited_set:
            visited_set.add(current)

            # If it's a terminal node, skip it
            if isinstance(current, Terminal):
                continue

            # Process non-terminal nodes in reverse order to maintain the depth-first order
            for i in reversed(range(len(current.values))):
                value = current.values[i]
                if isinstance(value, ModelVariable):
                    if allowed_vars is not None and value.name not in allowed_vars:
                        raise Exception(f"Invalid model variable name: {value.name}")
                    # Replace the ModelVariable with its value from 'model' (or the tokenizer if model does not have it)
                    # note we skip over attrs we don't have since we may be run twice, once on the model and once for the engine
                    if hasattr(model, value.name):
                        obj = model
                    elif hasattr(model, "tokenizer") and hasattr(
                        model.tokenizer, value.name
                    ):
                        obj = model.tokenizer
                    else:
                        obj = None
                    if obj is not None:
                        replacement_value = _wrap_as_grammar(getattr(obj, value.name))
                        if value.commit_point:
                            replacement_value = commit_point(
                                replacement_value, hidden=value.hidden
                            )
                        replacements.append(
                            (current, i, value)
                        )  # Record the replacement
                        current.values[i] = replacement_value  # Perform the replacement
                else:
                    # If not ModelVariable, push onto the stack to process later
                    stack.append((value, current, i))

    return replacements


# def replace_model_variables(grammar, model, visited_set={}):
#     '''Replace all the ModelVariable nodes with their values.'''

#     # see if we have already visited this node
#     if grammar in visited_set:
#         return []
#     else:
#         visited_set[grammar] = True

#     # we are done if this is a terminal
#     if isinstance(grammar, Terminal):
#         return []

#     # replace all matching sub-nodes
#     replacements = []
#     for i,value in enumerate(grammar.values):
#         if isinstance(value, ModelVariable):
#             g = _wrap_as_grammar(getattr(model, value.name))
#             if value.commit_point:
#                 g = commit_point(g, hidden=value.hidden)
#             replacements.append((grammar, i, value))
#             grammar.values[i] = g
#         else:
#             replacements.extend(replace_model_variables(value, model, visited_set))
#     return replacements


def unreplace_model_variables(replacements):
    """This restores a grammar back to its original state, ready for another execution."""
    for grammar, i, orig_value in replacements:
        grammar.values[i] = orig_value


def _wrap_as_grammar(value):
    """This takes whatever value was given and tries to turn in into a guidance grammar."""

    # if it is already a valid grammar we have no need to wrap it
    if isinstance(value, GrammarFunction):
        return value

    # if it is already a valid grammar we have no need to wrap it
    if value is None:
        return Null()

    # we have a constant value
    if isinstance(value, (str, bytes)):
        return string(value)

    raise Exception("Can't wrap as a grammar!")


def commit_point(value, hidden=False):
    """Force the grammar to commit to a parse that includes this node once it can.

    Not that commit point nodes can be optionally hidden (in fact they are the only
    nodes that can be hidden since they are by definition not impacted by multiple possible
    inconsistent parses.)"""
    # TODO: assert that value is not empty since we don't yet support that
    if isinstance(value, str):
        value = string(value)
    if isinstance(value, Terminal):
        value = Join(
            [value]
        )  # commit points should be full nodes (otherwise we can't hide them) TODO: decide if we want to do this even for non-hidden commit points
    value.commit_point = True
    if hidden:
        _rec_hide(value)
    return value


def _rec_hide(grammar):
    if not grammar.hidden:
        grammar.hidden = True
        if hasattr(grammar, "values"):
            for g in grammar.values:
                _rec_hide(g)


class Placeholder(GrammarFunction):
    __slots__ = tuple("nullable")

    def __init__(self):
        self.nullable = False


class Join(GrammarFunction):
    __slots__ = (
        "nullable",
        "values",
        "name",
        "hidden",
        "commit_point",
        "capture_name",
        "max_tokens",
    )

    def __init__(
        self, values, name: Union[str, None] = None, max_tokens=100000000
    ) -> None:
        values = [
            string(v) if isinstance(v, (str, bytes)) else v for v in values
        ]  # wrap raw strings
        self.nullable = all(getattr(v, "nullable", False) for v in values)
        self.values = [v for v in values if not isinstance(v, Null)]
        self.name = name if name is not None else GrammarFunction._new_name()
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.max_tokens = max_tokens

    def __repr__(self, indent="", done=None):
        if done is None:
            done = set()
        s = self.name.ljust(20) + " <- " + " ".join([v.name for v in self.values])
        s += (
            "        "
            + ("hidden " if self.hidden else "")
            + ("commit_point " if self.commit_point else "")
            + (f"capture_name={self.capture_name} " if self.capture_name else "")
            + (f"max_tokens={self.max_tokens}" if self.max_tokens < 100000 else "")
            + "\n"
        )
        done.add(self)
        for v in self.values:
            if v not in done and (isinstance(v, Join) or isinstance(v, Select)):
                s += v.__repr__(indent, done)
        return s

    def _to_proto(self, index_map):
        data = _serialization_pb2.Join()
        data.nullable = self.nullable
        for v in self.values:
            data.values.append(index_map[v])
        data.name = self.name
        data.hidden = self.hidden
        data.commit_point = self.commit_point
        data.capture_name = "" if self.capture_name is None else self.capture_name
        data.max_tokens = self.max_tokens
        return data

    @staticmethod
    def _from_proto(data):
        out = Join(
            data.values,  # we put ints in that will be replaced later by the deserialize method
            name=data.name,
            max_tokens=data.max_tokens,
        )
        out.nullable = data.nullable
        out.hidden = data.hidden
        out.commit_point = data.commit_point
        out.capture_name = None if data.capture_name == "" else data.capture_name
        return out


class Select(GrammarFunction):
    __slots__ = (
        "nullable",
        "_values",
        "name",
        "hidden",
        "commit_point",
        "capture_name",
        "max_tokens",
        "recursive",
    )

    def __init__(
        self, values, capture_name=None, name=None, max_tokens=10000000, recursive=False
    ) -> None:
        self.values = values
        self.name = name if name is not None else GrammarFunction._new_name()
        self.hidden = False
        self.commit_point = False
        self.capture_name = capture_name
        self.max_tokens = max_tokens
        self.recursive = recursive

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, vals):
        self._values = [string(v) if isinstance(v, (str, bytes)) else v for v in vals]
        self.nullable = any(getattr(v, "nullable", False) for v in self._values)
        self._values = [v for v in self._values if not isinstance(v, Null)]

    def __repr__(self, indent="", done=None):
        if done is None:
            done = set()
        s = self.name.ljust(20) + " <- " + " | ".join([v.name for v in self.values])
        s += (
            "        "
            + ("hidden " if self.hidden else "")
            + ("commit_point " if self.commit_point else "")
            + (f"max_tokens={self.max_tokens}" if self.max_tokens < 100000 else "")
            + "\n"
        )
        done.add(self)
        for v in self.values:
            if v not in done and (isinstance(v, Join) or isinstance(v, Select)):
                s += v.__repr__(indent, done)
        return s

    def _to_proto(self, index_map):
        data = _serialization_pb2.Select()
        data.nullable = self.nullable
        for v in self.values:
            data.values.append(index_map[v])
        data.name = self.name
        data.hidden = self.hidden
        data.commit_point = self.commit_point
        data.capture_name = "" if self.capture_name is None else self.capture_name
        data.max_tokens = self.max_tokens
        data.recursive = self.recursive

        return data

    @staticmethod
    def _from_proto(data):
        out = Select(
            data.values,  # we put ints in that will be replaced later by the deserialize method
            name=data.name,
            max_tokens=data.max_tokens,
        )
        out.nullable = data.nullable
        out.hidden = data.hidden
        out.commit_point = data.commit_point
        out.capture_name = None if data.capture_name == "" else data.capture_name
        out.recursive = data.recursive
        return out


def string(value: Union[str, bytes]) -> Union[Null, Byte, Join]:
    if isinstance(value, str):
        b = bytes(value, encoding="utf8")
    elif isinstance(value, bytes):
        b = value
    else:
        raise Exception("Must pass bytes or str to the string() function!")
    if len(value) == 0:
        return Null()
    elif len(b) == 1:
        return Byte(b)
    else:
        return Join([Byte(b[i : i + 1]) for i in range(len(b))], name=str(b))


def select(
    options: List[_T], name=None, list_append=False, recurse=False, skip_checks=False
) -> Union[Select, _T]:
    """Choose between a set of options.

    This function constrains the next generation from the LLM to be one of the
    given `options`.
    If the list only has a single element, then that value can be returned
    immediately, without calling the LLM.

    Parameters
    ----------
    name : str or None
        If this is not None then the the results of the generation will be saved as a variable on
        the Model object (so you can access the result as `lm["var_name"]`).

    options : List
        The set of available choices for the next generation

    list_append : bool
        If this is True then the results saved to `lm[name]` will not be written directly but rather appended
        to a list (if no list with the current name is present one will be created). This is useful for
        building lists inside python loops.

    recurse : bool
        Indicate whether multiple choices should be made. This is useful for tasks such as
        building up integers digit by digit: `select(options=list(range(10)), recurse=True)`

    skip_checks: bool
        Whether or not to perform sanity checks on the supplied options, to ensure
        that more obscure errors do not appear later.
    """
    # TODO: allow for returning the probabilites of the selected item
    # TODO: also the full probabilites distribution over all items. We can implement this using the prob of the selected item by repeating the call, removing the selected item each time
    if not skip_checks:
        for i, value in enumerate(options):
            assert not isinstance(
                value, RawFunction
            ), "You cannot select between stateful functions in the current guidance implementation!"
            assert not isinstance(
                value, types.FunctionType
            ), "Did you pass a function without calling it to select? You need to pass the results of a called guidance function to select."
            if isinstance(value, int) or isinstance(value, float):
                options[i] = str(value)  # type: ignore[assignment]

    # set up list append var saving if requested
    if list_append:
        name = "__LIST_APPEND:" + name

    if recurse:
        node = Select([], capture_name=name, recursive=True)
        if "" in options:
            # if we have an empty option, 'node + v' also covers the case of 'v' itself
            # thus, we don't have to add 'options' (except for the empty string)
            node.values = [node + v for v in options if v != ""] + [""]
        else:
            node.values = [node + v for v in options if v != ""] + options
        return node
    else:
        if len(options) == 1 and name is None:
            return options[0]
        else:
            return Select(options, capture_name=name, recursive=False)


def byte_range(low, high) -> ByteRange:
    return ByteRange(low + high)


# def ignore_placeholders(value):
#     if not isinstance(value, Join): # don't double wrap
#         value = Join([value]) # this ensures we capture what we want, and not something surprisingly self_recursive
#     value.ignore_placeholders = True
#     return value


def capture(value, name):
    # if log_probs:
    #     name += ":__LOG_PROBS"
    if not (isinstance(value, Join) and len(value.values) == 1):  # don't double wrap
        value = Join(
            [value]
        )  # this ensures we capture what we want, and not something surprisingly self_recursive
    value.capture_name = name
    return value


def token_limit(value, max_tokens: int):
    _rec_token_limit(value, max_tokens)
    return value


def _rec_token_limit(grammar, max_tokens: int):
    if grammar.max_tokens > max_tokens and not isinstance(grammar, Terminal):
        if getattr(
            grammar, "recursive", False
        ):  # only restrict recursive selects, otherwise we would block all ways to complete the grammar
            grammar.max_tokens = max_tokens
            for value in getattr(
                grammar, "values", []
            ):  # restrict recursive selects recursive nodes
                if not isinstance(value, Terminal):
                    value.max_tokens = max_tokens
        if hasattr(grammar, "values"):
            for g in grammar.values:
                _rec_token_limit(g, max_tokens)


def with_temperature(value, temperature):
    """This sets the sampling temperature to be used for the given portion of the grammar.

    Note that if the grammar passed to us already has some portions with a temperature
    setting in place, those settings will not be overridden.
    """
    _re_with_temperature(value, temperature, {})
    return value


def _re_with_temperature(grammar, temperature, visited_set):

    # don't go down the same path twice
    if grammar in visited_set:
        return
    visited_set[grammar] = True

    # if getattr(grammar, "temperature", 100000000) > temperature:
    if (
        isinstance(grammar, Terminal) and grammar.temperature < 0
    ):  # only need to set temp for terminals
        grammar.temperature = temperature
    elif getattr(grammar, "temperature", 100000000) > temperature and hasattr(
        grammar, "values"
    ):
        for g in grammar.values:
            _re_with_temperature(g, temperature, visited_set)


# def model_variable(name):
#     return ModelVariable(name)


def active_role_end() -> ModelVariable:
    return ModelVariable("active_role_end")


def eos_token() -> ModelVariable:
    return ModelVariable("eos_token")


def bos_token() -> ModelVariable:
    return ModelVariable("bos_token")


_null_grammar = string("")


# def char_range(low, high):
#     low_bytes = bytes(low, encoding="utf8")
#     high_bytes = bytes(high, encoding="utf8")
#     if len(low_bytes) > 1 or len(high_bytes) > 1:
#         raise Exception("We don't yet support multi-byte character ranges!")
#     return ByteRange(low_bytes + high_bytes)
def str_to_grammar(value: str):
    is_id = False
    parts = re.split(_tag_pattern, value)

    # we have no embedded objects
    if len(parts) == 1:
        return string(value)

    # if we have embedded objects we have to convert the string to a grammar tree
    else:
        partial_grammar: Any = _null_grammar
        # lm.suffix = ""
        for i, part in enumerate(parts):
            # if i < len(parts) - 1:
            #     lm.suffix = parts[i+1]
            if is_id:
                call = _call_pool[part]
                if isinstance(call, GrammarFunction):
                    partial_grammar += _call_pool[part]
                else:
                    partial_grammar = RawFunction(
                        lambda lm, g, call: call(lm + g),
                        partial_grammar,
                        _call_pool[part],
                    )
                    # lm += partial_grammar
                    # lm = _call_pool[part](lm)
                    # partial_grammar = _null_grammar
            elif part != "":
                partial_grammar += string(part)
            is_id = not is_id
    return partial_grammar



import re
from .._model import Chat, Instruct
from .._grammarless import GrammarlessEngine, Grammarless

try:
    import vertexai

    is_vertexai = True
except ModuleNotFoundError:
    is_vertexai = False


class VertexAIEngine(GrammarlessEngine):
    def __init__(
        self, tokenizer, max_streaming_tokens, timeout, compute_log_probs, model_obj
    ):
        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)
        self.model_obj = model_obj


class VertexAI(Grammarless):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        max_streaming_tokens=None,
        timeout=0.5,
        compute_log_probs=False,
        engine_class=None,
        **kwargs,
    ):
        """Build a new VertexAI model object that represents a model in a given state."""
        if not is_vertexai:
            raise Exception(
                "Please install the vertexai package using `pip install google-cloud-aiplatform` in order to use guidance.models.VertexAI!"
            )

        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is VertexAI:
            found_subclass = None
            from .. import vertexai as vertexai_subclasses

            if isinstance(model, str):
                model_name = model
            else:
                model_name = model._model_id

            # CodeyCompletion
            if re.match("code-gecko(@[0-9]+)?", model_name):
                found_subclass = vertexai_subclasses.CodeyCompletion

            # CodeyInstruct
            elif re.match("code-bison(@[0-9]+)?", model_name):
                found_subclass = vertexai_subclasses.CodeyInstruct

            # CodeyChat
            elif re.match("codechat-bison(@[0-9]+)?", model_name):
                found_subclass = vertexai_subclasses.CodeyChat

            # PaLM2Instruct
            elif re.match("text-(bison|unicorn)(@[0-9]+)?", model_name):
                found_subclass = vertexai_subclasses.PaLM2Instruct

            # PaLM2Chat
            elif re.match("chat-bison(@[0-9]+)?", model_name):
                found_subclass = vertexai_subclasses.PaLM2Chat

            # Gemini2Chat
            elif re.match("gemini-(pro|ultra)(@[0-9]+)?", model_name):
                found_subclass = vertexai_subclasses.GeminiChat

            # convert to any found subclass
            if found_subclass is not None:
                self.__class__ = found_subclass
                found_subclass.__init__(
                    self,
                    model,
                    tokenizer=tokenizer,
                    echo=echo,
                    max_streaming_tokens=max_streaming_tokens,
                    **kwargs,
                )
                return  # we return since we just ran init above and don't need to run again

            # make sure we have a valid model object
            if isinstance(model, str):
                raise Exception(
                    "The model ID you passed, `{model}`, does not match any known subclasses!"
                )

        # this allows us to use a single constructor for all our subclasses
        if engine_class is None:
            engine_map = {
                VertexAICompletion: VertexAICompletionEngine,
                VertexAIInstruct: VertexAIInstructEngine,
                VertexAIChat: VertexAIChatEngine,
            }
            for k in engine_map:
                if issubclass(self.__class__, k):
                    engine_class = engine_map[k]
                    break

        super().__init__(
            engine_class(
                tokenizer=tokenizer,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                max_streaming_tokens=max_streaming_tokens,
                model_obj=model,
            ),
            echo=echo,
        )


class VertexAICompletion(VertexAI):
    pass


class VertexAICompletionEngine(VertexAIEngine):

    def _generator(self, prompt, temperature):
        self._not_running_stream.clear()  # so we know we are running
        self._data = prompt  # we start with this data

        try:
            kwargs = {}
            if self.max_streaming_tokens is not None:
                kwargs["max_output_tokens"] = self.max_streaming_tokens
            generator = self.model_obj.predict_streaming(
                prompt.decode("utf8"),
                # top_p=self.top_p,
                temperature=temperature,
                **kwargs,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for chunk in generator:
            yield chunk.text.encode("utf8")


class VertexAIInstruct(VertexAI, Instruct):

    def get_role_start(self, name):
        return ""

    def get_role_end(self, name):
        if name == "instruction":
            return "<|endofprompt|>"
        else:
            raise Exception(
                f"The VertexAIInstruct model does not know about the {name} role type!"
            )


class VertexAIInstructEngine(VertexAIEngine):
    def _generator(self, prompt, temperature):
        # start the new stream
        prompt_end = prompt.find(b"<|endofprompt|>")
        if prompt_end >= 0:
            stripped_prompt = prompt[:prompt_end]
        else:
            raise Exception(
                "This model cannot handle prompts that don't match the instruct format! Follow for example:\nwith instruction():\n    lm += prompt\nlm += gen(max_tokens=10)"
            )
        self._not_running_stream.clear()  # so we know we are running
        self._data = stripped_prompt + b"<|endofprompt|>"  # we start with this data
        kwargs = {}
        if self.max_streaming_tokens is not None:
            kwargs["max_output_tokens"] = self.max_streaming_tokens
        for chunk in self.model_obj.predict_streaming(
            self._data.decode("utf8"), temperature=temperature, **kwargs
        ):
            yield chunk.text.encode("utf8")


class VertexAIChat(VertexAI, Chat):
    pass


class VertexAIChatEngine(VertexAIEngine):

    def _generator(self, prompt, temperature):

        # find the system text
        pos = 0
        system_start = b"<|im_start|>system\n"
        user_start = b"<|im_start|>user\n"
        assistant_start = b"<|im_start|>assistant\n"
        role_end = b"<|im_end|>"
        # system_start_pos = prompt.startswith(system_start)

        # find the system text
        system_text = b""
        if prompt.startswith(system_start):
            pos += len(system_start)
            system_end_pos = prompt.find(role_end)
            system_text = prompt[pos:system_end_pos]
            pos = system_end_pos + len(role_end)

        # find the user/assistant pairs
        messages = []
        valid_end = False
        while True:

            # find the user text
            if prompt[pos:].startswith(user_start):
                pos += len(user_start)
                end_pos = prompt[pos:].find(role_end)
                if end_pos < 0:
                    break
                messages.append(
                    dict(
                        role="user",
                        content=prompt[pos : pos + end_pos].decode("utf8"),
                    )
                )
                pos += end_pos + len(role_end)
            elif prompt[pos:].startswith(assistant_start):
                pos += len(assistant_start)
                end_pos = prompt[pos:].find(role_end)
                if end_pos < 0:
                    valid_end = True
                    break
                messages.append(
                    dict(
                        role="assistant",
                        content=prompt[pos : pos + end_pos].decode("utf8"),
                    )
                )
                pos += end_pos + len(role_end)
            else:
                raise Exception(
                    "It looks like your prompt is not a well formed chat prompt! Please enclose all model state appends inside chat role blocks like `user()` or `assistant()`."
                )

        self._data = prompt[:pos]

        assert len(messages) > 0, "Bad chat format! No chat blocks were defined."
        assert (
            messages[-1]["role"] == "user"
        ), "Bad chat format! There must be a user() role before the last assistant() role."
        assert valid_end, "Bad chat format! You must generate inside assistant() roles."

        # TODO: don't make a new session on every call
        # last_user_text = messages.pop().content

        return self._start_generator(system_text.decode("utf8"), messages, temperature)

        # kwargs = {}
        # if self.max_streaming_tokens is not None:
        #     kwargs["max_output_tokens"] = self.max_streaming_tokens
        # generator = chat_session.send_message_streaming(last_user_text, temperature=temperature, **kwargs)

        # for chunk in generator:
        #     yield chunk.text.encode("utf8")

    def _start_generator(self, system_text, messages, temperature):
        messages = [
            vertexai.language_models.ChatMessage(author=m["role"], content=m["content"])
            for m in messages
        ]
        last_user_text = messages.pop().content

        chat_session = self.model_obj.start_chat(
            context=system_text,
            message_history=messages,
        )

        kwargs = {}
        if self.max_streaming_tokens is not None:
            kwargs["max_output_tokens"] = self.max_streaming_tokens
        generator = chat_session.send_message_streaming(
            last_user_text, temperature=temperature, **kwargs
        )

        for chunk in generator:
            yield chunk.text.encode("utf8")



import json
import multiprocessing
import time
from typing import List, Union

import pytest
import requests
from jsonschema import validate

from guidance import Server, gen, models
from guidance.library import json as gen_json
from guidance.library._json import _to_compact_json

# Everything in here is a 'server' test
# Mark is configured in pyproject.toml
pytestmark = pytest.mark.server

# We spin out a separate process, and it
# has to start up and get ready to
# respond to requests. Just waiting is
# not ideal, but is the simplest option
PROCESS_DELAY_SECS = 40


def server_process(*, mock_string: Union[str, List[str]] = ""):
    byte_patterns = []
    if isinstance(mock_string, str):
        byte_patterns = [f"<s>{mock_string}".encode()]
    else:
        for s in mock_string:
            byte_patterns.append(f"<s>{s}".encode())
    lm = models.Mock(byte_patterns=byte_patterns)

    temp_lm = lm + gen()
    print(f"=====Plain gen output: {temp_lm}=====")

    server = Server(lm, api_key="SDFSDF")
    server.run(port=8392)


class ServerContext:
    def __init__(self, mock_string: str):
        self._process = multiprocessing.Process(
            target=server_process, kwargs=dict(mock_string=mock_string)
        )

    def __enter__(self):
        self._process.start()
        time.sleep(PROCESS_DELAY_SECS)
        return None

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._process.terminate()
        time.sleep(PROCESS_DELAY_SECS)
        return False  # We don't handle exceptions


def test_remote_mock_gen():
    with ServerContext(mock_string=""):
        m = models.Model("http://localhost:8392", api_key="SDFSDF")
        m2 = m + "A story." + gen("test", max_tokens=20)
        assert len(str(m2)) > 20, "The model didn't generate enough data"


def test_remote_mock_gen_bad_auth():
    with ServerContext(mock_string=""):
        m = models.Model("http://localhost:8392", api_key="FDSFDS")

        with pytest.raises(requests.exceptions.HTTPError) as http_err:
            _ = m + "A story." + gen("test", max_tokens=20)
        assert http_err.value.response.status_code == 401
        assert http_err.value.response.text == '{"detail":"Invalid API key"}'


def test_return_mock_string():
    my_string = "My roundtrip"
    with ServerContext(mock_string=my_string):
        m = models.Model("http://localhost:8392", api_key="SDFSDF")
        m2 = m + gen(max_tokens=10, name="captured")
        assert m2["captured"].startswith(my_string)


@pytest.mark.parametrize(
    "target_obj",
    [
        dict(my_list=None),
        dict(my_list=dict(my_str="a", next=None)),
        dict(my_list=dict(my_str="a", next=dict(my_str="b", next=None))),
    ],
)
def test_remote_gen_json(target_obj):
    schema = """
{
    "$defs": {
        "A": {
            "properties": {
                "my_str": {
                    "default": "me",
                    "title": "My Str",
                    "type": "string"
                },
                "next": {
                    "anyOf": [
                        {
                            "$ref": "#/$defs/A"
                        },
                        {
                            "type": "null"
                        }
                    ]
                }
            },
            "type": "object"
        }
    },
    "type": "object",
    "properties": {
        "my_list": {
            "anyOf": [
                {
                    "$ref": "#/$defs/A"
                },
                {
                    "type": "null"
                }
            ]
        }
    }
}
        """
    schema_obj = json.loads(schema)

    # Sanity check input
    validate(target_obj, schema_obj)

    print(f"target_obj={_to_compact_json(target_obj)}")

    with ServerContext(mock_string=[_to_compact_json(target_obj)]):
        m = models.Model("http://localhost:8392", api_key="SDFSDF")
        m += gen_json(schema=schema_obj, name="my_json_string")
        print(f"Raw: {m['my_json_string']}")

        my_obj = json.loads(m["my_json_string"])
        print(f"Received object: {json.dumps(my_obj, indent=4)}")
        validate(my_obj, schema_obj)
        assert my_obj == target_obj



import os
import tiktoken

from ._model import Chat, Instruct
from ._grammarless import GrammarlessEngine, Grammarless


class AnthropicEngine(GrammarlessEngine):
    def __init__(
        self,
        model,
        tokenizer,
        api_key,
        timeout,
        max_streaming_tokens,
        compute_log_probs,
        **kwargs,
    ):
        try:
            from anthropic import Anthropic
        except ModuleNotFoundError:
            raise Exception(
                "Please install the anthropic package version >= 0.7 using `pip install anthropic -U` in order to use guidance.models.Anthropic!"
            )

        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is Anthropic:
            raise Exception(
                "The Anthropic class is not meant to be used directly! Please use AnthropicChat assuming the model you are using is chat-based."
            )

        if api_key is None:
            api_key = os.environ.get("ANTHROPIC_API_KEY")

        if api_key is None:
            raise Exception(
                "Expected an api_key argument or the ANTHROPIC_API_KEY environment variable to be set!"
            )

        self.anthropic = Anthropic(api_key=api_key, **kwargs)

        self.model_name = model

        # we pretend it tokenizes like gpt2 if tiktoken does not know about it... TODO: make this better
        if tokenizer is None:
            try:
                tokenizer = tiktoken.encoding_for_model(model)
            except:
                tokenizer = tiktoken.get_encoding("gpt2")

        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)

    def _generator(self, prompt, temperature):

        # update our shared data state
        self._reset_shared_data(prompt, temperature)

        try:
            generator = self.anthropic.completions.create(
                model=self.model_name,
                prompt=prompt.decode("utf8"),
                max_tokens_to_sample=self.max_streaming_tokens,
                stream=True,
                temperature=temperature,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for part in generator:
            chunk = part.completion or ""
            # print(chunk)
            yield chunk.encode("utf8")


class Anthropic(Grammarless):
    """Represents an Anthropic model as exposed through their remote API.

    Note that because this uses a remote API endpoint without built-in guidance support
    there are some things we cannot do, like force the model to follow a pattern inside
    a chat role block.
    """

    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        api_key=None,
        timeout=0.5,
        max_streaming_tokens=1000,
        compute_log_probs=False,
        **kwargs,
    ):
        """Build a new Anthropic model object that represents a model in a given state."""

        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is Anthropic:
            found_subclass = None

            # chat
            found_subclass = AnthropicChat  # we assume all models are chat right now

            # convert to any found subclass
            self.__class__ = found_subclass
            found_subclass.__init__(
                self,
                model,
                tokenizer=tokenizer,
                echo=echo,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                **kwargs,
            )
            return  # we return since we just ran init above and don't need to run again

        super().__init__(
            engine=AnthropicEngine(
                model=model,
                tokenizer=tokenizer,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                **kwargs,
            ),
            echo=echo,
        )


class AnthropicChat(Anthropic, Chat):
    def get_role_start(self, role_name, **kwargs):
        if role_name == "user":
            return "\n\nHuman:"
        if role_name == "assistant":
            return "\n\nAssistant:"
        if role_name == "system":
            return ""

    def get_role_end(self, role_name=None):
        return ""



__version__ = "0.1.14"

import sys
import types

from . import models
from ._guidance import _decorator, guidance

from ._grammar import (
    Placeholder,
    RawFunction,
    GrammarFunction,
    Terminal,
    replace_grammar_node,
    string,
)
from ._utils import strip_multiline_string_indents
from ._server import Server


# This makes the guidance module callable
class _Guidance(types.ModuleType):
    def __call__(
        self, f=None, *, stateless=False, cache=None, dedent=True, model=models.Model
    ):
        return _decorator(
            f, stateless=stateless, cache=cache, dedent=dedent, model=model
        )


sys.modules[__name__].__class__ = _Guidance

# we expose all the library functions at the top level of the module
from .library import *



from ._transformers import Transformers, TransformersChat


class Llama(Transformers):
    pass


class LlamaChat(TransformersChat, Llama):

    def system(self):
        """Patch up the system command to convert normal system role structure into Llama structure (nested in the first user message)."""
        self._system_prefex = "[INST] " if str(self) == "" else ""
        out = super().system()
        delattr(self, "_system_prefex")
        return out

    def get_role_start(self, role_name, **kwargs):
        if role_name == "system":
            return self._system_prefex + "<<SYS>>\n"
        elif role_name == "user":
            if str(self).endswith("\n<</SYS>>\n\n"):
                return ""  # we don't need to start anything if we are starting with a top level unnested system tag
            else:
                return "[INST] "
        else:
            return " "

    def get_role_end(self, role_name=None):
        if role_name == "system":
            return "\n<</SYS>>\n\n"
        elif role_name == "user":
            return " [/INST]"
        else:
            return " "



from .._grammar import byte_range


def char_range(low, high):
    low_bytes = bytes(low, encoding="utf8")
    high_bytes = bytes(high, encoding="utf8")
    if len(low_bytes) > 1 or len(high_bytes) > 1:
        raise Exception("We don't yet support multi-byte character ranges!")
    return byte_range(low_bytes, high_bytes)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: _serialization.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x14_serialization.proto\x12\x08guidance"3\n\x07Grammar\x12(\n\x05nodes\x18\x01 \x03(\x0b\x32\x19.guidance.GrammarFunction"\x83\x03\n\x12\x45ngineCallResponse\x12\x11\n\tnew_bytes\x18\x01 \x01(\x0c\x12\x14\n\x0cis_generated\x18\x02 \x01(\x08\x12\x16\n\x0enew_bytes_prob\x18\x03 \x01(\x02\x12G\n\x0e\x63\x61pture_groups\x18\x04 \x03(\x0b\x32/.guidance.EngineCallResponse.CaptureGroupsEntry\x12W\n\x17\x63\x61pture_group_log_probs\x18\x05 \x03(\x0b\x32\x36.guidance.EngineCallResponse.CaptureGroupLogProbsEntry\x12\x17\n\x0fnew_token_count\x18\x06 \x01(\x05\x1a\x34\n\x12\x43\x61ptureGroupsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a;\n\x19\x43\x61ptureGroupLogProbsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01"w\n\x04\x42yte\x12\x0c\n\x04\x62yte\x18\x01 \x01(\x0c\x12\x0e\n\x06hidden\x18\x02 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x03 \x01(\x08\x12\x10\n\x08nullable\x18\x04 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x05 \x01(\t\x12\x13\n\x0btemperature\x18\x06 \x01(\x02"p\n\tByteRange\x12\x12\n\nbyte_range\x18\x01 \x01(\x0c\x12\x0e\n\x06hidden\x18\x03 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x04 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x05 \x01(\t\x12\x13\n\x0btemperature\x18\x06 \x01(\x02"\x06\n\x04Null"k\n\rModelVariable\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06hidden\x18\x02 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x03 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x04 \x01(\t\x12\x10\n\x08nullable\x18\x05 \x01(\x08"\x86\x01\n\x04Join\x12\x10\n\x08nullable\x18\x01 \x01(\x08\x12\x0e\n\x06values\x18\x02 \x03(\x05\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06hidden\x18\x04 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x05 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x06 \x01(\t\x12\x12\n\nmax_tokens\x18\x07 \x01(\x05"\x9b\x01\n\x06Select\x12\x10\n\x08nullable\x18\x01 \x01(\x08\x12\x0e\n\x06values\x18\x02 \x03(\x05\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06hidden\x18\x04 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x05 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x06 \x01(\t\x12\x12\n\nmax_tokens\x18\x07 \x01(\x05\x12\x11\n\trecursive\x18\x08 \x01(\x08"\xe4\x01\n\x0fGrammarFunction\x12\x1e\n\x04join\x18\x01 \x01(\x0b\x32\x0e.guidance.JoinH\x00\x12"\n\x06select\x18\x02 \x01(\x0b\x32\x10.guidance.SelectH\x00\x12\x1e\n\x04\x62yte\x18\x03 \x01(\x0b\x32\x0e.guidance.ByteH\x00\x12)\n\nbyte_range\x18\x04 \x01(\x0b\x32\x13.guidance.ByteRangeH\x00\x12\x31\n\x0emodel_variable\x18\x05 \x01(\x0b\x32\x17.guidance.ModelVariableH\x00\x42\x0f\n\rfunction_typeb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "_serialization_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY"]._options = None
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY"]._serialized_options = b"8\001"
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY"]._options = None
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY"]._serialized_options = (
        b"8\001"
    )
    _globals["_GRAMMAR"]._serialized_start = 34
    _globals["_GRAMMAR"]._serialized_end = 85
    _globals["_ENGINECALLRESPONSE"]._serialized_start = 88
    _globals["_ENGINECALLRESPONSE"]._serialized_end = 475
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY"]._serialized_start = 362
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY"]._serialized_end = 414
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY"]._serialized_start = 416
    _globals["_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY"]._serialized_end = 475
    _globals["_BYTE"]._serialized_start = 477
    _globals["_BYTE"]._serialized_end = 596
    _globals["_BYTERANGE"]._serialized_start = 598
    _globals["_BYTERANGE"]._serialized_end = 710
    _globals["_NULL"]._serialized_start = 712
    _globals["_NULL"]._serialized_end = 718
    _globals["_MODELVARIABLE"]._serialized_start = 720
    _globals["_MODELVARIABLE"]._serialized_end = 827
    _globals["_JOIN"]._serialized_start = 830
    _globals["_JOIN"]._serialized_end = 964
    _globals["_SELECT"]._serialized_start = 967
    _globals["_SELECT"]._serialized_end = 1122
    _globals["_GRAMMARFUNCTION"]._serialized_start = 1125
    _globals["_GRAMMARFUNCTION"]._serialized_end = 1353
# @@protoc_insertion_point(module_scope)



import os
import atexit
from pathlib import Path
from itertools import takewhile
import operator
import sys
import logging
import numpy as np

from .._model import Tokenizer, Engine, Model, Chat
from .._remote import RemoteEngine
from ..._utils import normalize_notebook_stdout_stderr

try:
    import llama_cpp

    is_llama_cpp = True
except ModuleNotFoundError:
    is_llama_cpp = False

logger = logging.getLogger(__name__)

shutdown_none = True


def set_shutdown_flag():
    global shutdown_none
    # python can set anything to None at shutdown, so use None
    shutdown_none = None


atexit.register(set_shutdown_flag)


class _LlamaBatchContext:
    def __init__(self, n_batch, n_ctx):
        self._llama_batch_free = llama_cpp.llama_batch_free
        self.batch = llama_cpp.llama_batch_init(n_batch, 0, n_ctx)
        if self.batch is None:
            raise Exception("call to llama_cpp.llama_batch_init returned NULL.")

    def __del__(self):
        if shutdown_none is not None:
            llama_batch_free = getattr(self, "_llama_batch_free", None)
            batch = getattr(self, "batch", None)
            if batch is not None and llama_batch_free is not None:
                self._llama_batch_free = None
                self.batch = None
                llama_batch_free(batch)


class LlamaCppTokenizer(Tokenizer):
    def __init__(self, model_obj):
        self._model_obj = model_obj

        tokenizer = llama_cpp.LlamaTokenizer(model_obj)
        if not hasattr(tokenizer, "llama"):
            tokenizer.llama = tokenizer._model

        # get the bytes strings for all the tokens
        tokens = []
        for i in range(tokenizer.llama.n_vocab()):
            tok = tokenizer.llama.detokenize(
                [i]
            )  # note that detokenize returns bytes directly
            if tok == b"":
                tok = llama_cpp.llama_token_get_text(
                    model_obj.model, i
                )  # get text rep of special tokens
            tokens.append(tok)

        super().__init__(
            tokens, tokenizer.llama.token_bos(), tokenizer.llama.token_eos()
        )

    def __call__(self, byte_string):
        return self._model_obj.tokenize(byte_string, add_bos=False, special=True)


class LlamaCppEngine(Engine):
    """The core class that runs inference using llama.cpp."""

    def __init__(self, model, compute_log_probs, **kwargs):
        if not is_llama_cpp:
            raise Exception(
                "Please install llama-cpp-python with `pip install llama-cpp-python` in order to use guidance.models.LlamaCpp!"
            )

        if isinstance(model, Path):
            model = str(model)
        if model is None or isinstance(model, str) and len(model.strip()) == 0:
            model = os.environ.get("LLAMA_CPP_MODEL", "")
            if len(model.strip()) == 0:
                try:
                    with open(os.path.expanduser("~/.llama_cpp_model"), "r") as file:
                        model = file.read().replace("\n", "")
                except:
                    pass
                if len(model.strip()) == 0:
                    raise ValueError(
                        "If model is None then a model file must be specified in either the LLAMA_CPP_MODEL environment variable or in the ~/.llama_cpp_model file."
                    )

        if isinstance(model, str):
            self.model = model
            if "verbose" not in kwargs:
                kwargs["verbose"] = False

            # patch over https://github.com/abetlen/llama-cpp-python/issues/729
            try:
                sys.stdout.fileno()
            except:
                logger.warn(
                    "Cannot use verbose=True in this context (probably CoLab). See https://github.com/abetlen/llama-cpp-python/issues/729"
                )
                kwargs["verbose"] = (
                    True  # llama-cpp-python can't hide output in this case
                )

            with normalize_notebook_stdout_stderr():
                self.model_obj = llama_cpp.Llama(model_path=model, logits_all=True, **kwargs)
        elif isinstance(model, llama_cpp.Llama):
            self.model = model.__class__.__name__
            self.model_obj = model
        else:
            raise TypeError(
                "model must be None, a file path string, or a llama_cpp.Llama object."
            )

        self._context = _LlamaBatchContext(
            self.model_obj.n_batch, self.model_obj.n_ctx()
        )
        self._cache_token_ids = []

        super().__init__(
            LlamaCppTokenizer(self.model_obj), compute_log_probs=compute_log_probs
        )

        self._n_vocab = len(self.tokenizer.tokens)

    def _joint_tokenize(self, token_ids):
        """What a full joint tokenizer would give for a given byte string"""
        byte_string = b"".join([self.tokenizer.tokens[t] for t in token_ids])
        return self.tokenizer(byte_string)

    def get_logits(self, token_ids, forced_bytes, current_temp):
        """Computes the logits for the given token state.

        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        """

        if len(token_ids) == 0:
            raise ValueError("token_ids must contain some tokens.")

        # check what we have already cached
        cache_token_ids = self._cache_token_ids
        num_cached = sum(
            takewhile(operator.truth, map(operator.eq, token_ids, cache_token_ids))
        )
        if num_cached == len(token_ids):
            if num_cached == len(cache_token_ids):
                return self._cached_logits
            num_cached = (
                num_cached - 1
            )  # llama_cpp doesn't like it when we pass in 0 new tokens, so re-input one

        # make sure we don't run off the end of the model's context
        if self.model_obj.n_ctx() <= len(token_ids):
            raise Exception(
                f"Attempted to use a context length of {len(token_ids)} tokens, but this LlamaCpp model is only configured to support up to {self.model_obj.n_ctx()}!"
            )

        self._cache_token_ids = token_ids.copy()

        # clear obsolete parts of kv cache
        llama_cpp.llama_kv_cache_seq_rm(self.model_obj.ctx, -1, num_cached, -1)

        # eval the model
        n_batch = self.model_obj.n_batch
        batch = self._context.batch
        for i in range(num_cached, len(token_ids), n_batch):
            n_tokens = min(i + n_batch, len(token_ids)) - i
            batch.n_tokens = n_tokens
            for j in range(n_tokens):
                batch.token[j] = token_ids[i + j]
                batch.pos[j] = i + j
                batch.seq_id[j][0] = 0
                batch.n_seq_id[j] = 1
                batch.logits[j] = False

            if i + n_tokens == len(token_ids):
                batch.logits[n_tokens - 1] = True

            ret = llama_cpp.llama_decode(self.model_obj.ctx, batch)
            self.metrics.engine_input_tokens += n_tokens
            if ret != 0:
                raise Exception(f"Call to llama_cpp.llama_decode returned {ret}.")

        self.metrics.engine_output_tokens += 1

        # get the logits
        logits = llama_cpp.llama_get_logits(self.model_obj.ctx)
        if llama_cpp.__version__ < "0.2.58":
            logits = logits[(n_tokens - 1) * self._n_vocab : n_tokens * self._n_vocab]
        logits = np.ctypeslib.as_array(logits, shape=(self._n_vocab,)).copy()

        self._cached_logits = logits

        return logits


class LlamaCpp(Model):
    def __init__(
        self,
        model=None,
        echo=True,
        compute_log_probs=False,
        api_key=None,
        **llama_cpp_kwargs,
    ):
        """Build a new LlamaCpp model object that represents a model in a given state."""

        if isinstance(model, str) and model.startswith("http"):
            engine = RemoteEngine(model, api_key=api_key, **llama_cpp_kwargs)
        else:
            engine = LlamaCppEngine(
                model, compute_log_probs=compute_log_probs, **llama_cpp_kwargs
            )

        super().__init__(engine, echo=echo)


class LlamaCppChat(LlamaCpp, Chat):
    def get_role_start(self, role_name, **kwargs):
        if role_name == "user":

            # if we follow an auto-nested system role then we are done
            if self._current_prompt().endswith("\n<</SYS>>\n\n"):
                return ""
            else:
                return "[INST] "

        elif role_name == "assistant":
            return " "

        elif role_name == "system":

            # check if we are already embedded at the top of a user role
            if self._current_prompt().endswith("[INST] "):
                return "<<SYS>>\n"

            # if not then we auto nest ourselves
            else:
                return "[INST] <<SYS>>\n"

    def get_role_end(self, role_name=None):
        if role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return " "
        elif role_name == "system":
            return "\n<</SYS>>\n\n"



"""
Unit tests guidance._utils
"""

from typing import Tuple

import numpy as np
import pytest
import torch

from guidance import _utils


@pytest.fixture(scope="module")
def atol() -> float:
    return 1e-6


@pytest.mark.parametrize(
    "size_and_axis",
    [
        ((32_000,), -1),  # very next token logits
        ((10, 32_000), -1),  # many token's next-token logits
        ((4, 10, 32_000), -1),  # batch of texts
    ],
)
class TestLogitsTransforms:
    def test_log_softmax(self, size_and_axis, atol: float):
        size, axis = size_and_axis
        logits: np.ndarray = -np.random.uniform(low=0, high=60, size=size)
        log_probs = _utils.log_softmax(logits, axis=axis)
        log_probs_correct = torch.tensor(logits).log_softmax(dim=axis).numpy()
        assert np.allclose(log_probs, log_probs_correct, atol=atol)

    def test_softmax(self, size_and_axis, atol: float):
        size, axis = size_and_axis
        logits: np.ndarray = -np.random.uniform(low=0, high=60, size=size)
        probs = _utils.softmax(logits, axis=axis)
        probs_correct = torch.tensor(logits).softmax(dim=axis).numpy()
        assert np.allclose(probs, probs_correct, atol=atol)



from .._guidance import guidance
from ._block import block


@guidance
def set_attr_opener(lm, name, value):
    if hasattr(lm, name):
        lm = lm.setattr("__save" + name, getattr(lm, name))
    return lm.setattr(name, value)


@guidance
def set_attr_closer(lm, name):
    if hasattr(lm, "__save" + name):
        return lm.setattr(name, lm["__save" + name]).delattr("__save" + name)
    else:
        return lm.delattr(name)


def set_attribute(name, value=True):
    return block(
        opener=set_attr_opener(name, value),
        closer=set_attr_closer(name),
    )



import os
import re

try:
    import torch
except ModuleNotFoundError:
    pass

from .._model import Tokenizer, Engine, Model, Chat


class TransformersTokenizer(Tokenizer):
    def __init__(self, model, tokenizer, ignore_bos_token=False):
        if tokenizer is None:
            tokenizer = self._tokenizer(model)

        self._orig_tokenizer = tokenizer
        special_tokens_map = {
            id: token for token, id in tokenizer.get_added_vocab().items()
        }

        # build out the set of byte_string tokens
        byte_tokens = [None] * len(tokenizer)
        if hasattr(tokenizer, "byte_decoder"):
            byte_decoder = tokenizer.byte_decoder

            for i in range(len(tokenizer)):
                byte_coded = bytes(
                    [byte_decoder[c] for c in tokenizer.convert_ids_to_tokens(i)]
                )
                byte_tokens[i] = byte_coded

        elif hasattr(tokenizer, "sp_model"):
            space_prefix = "▁".encode()
            for i in range(len(tokenizer)):
                if i in special_tokens_map:
                    byte_coded = special_tokens_map[i].encode()
                else:
                    byte_coded = re.sub(
                        rb"<0x(..)>",
                        lambda x: bytes.fromhex(x[1].decode()),
                        tokenizer.sp_model.id_to_piece(i).encode(),
                    )
                byte_tokens[i] = byte_coded.replace(space_prefix, b" ")

        else:
            import transformers

            byte_decoder = transformers.AutoTokenizer.from_pretrained(
                "gpt2", use_fast=False
            ).byte_decoder  # fall back to gpt2 mapping

            # some special tokens may not have their whitespace encoded...
            byte_decoder[" "] = 32
            byte_decoder["\n"] = 10
            byte_decoder["\r"] = 13
            byte_decoder["\t"] = 9
            byte_decoder["▁"] = 32

            # run a quick spot check to verify we can rebuild complex multi-token unicode symbols
            s = "’•¶∂ƒ˙∆£Ħ爨ൠᅘ∰፨"
            t = tokenizer
            reconstructed = b""
            try:
                for i in t(s)["input_ids"]:
                    reconstructed += bytes(
                        [byte_decoder[c] for c in t.convert_ids_to_tokens(i)]
                    )
            except:
                raise ValueError(
                    f"The tokenizer being used is unable to convert a special character in {s}. For models with sentencepiece based tokenizers (e.g. llama, phi-3-mini), installing sentencepiece often fixes this issue (pip install sentencepiece)."
                )
            assert (
                reconstructed.decode() == s
            ), "The passed tokenizer does not have a byte_decoder property and using a standard gpt2 byte_decoder fails!"

            for i in range(len(tokenizer)):
                byte_coded = bytes(
                    [byte_decoder[c] for c in tokenizer.convert_ids_to_tokens(i)]
                )
                byte_tokens[i] = byte_coded

        # the superclass does most of the work once we have the tokens
        super().__init__(
            byte_tokens,
            None if ignore_bos_token else tokenizer.bos_token_id,
            tokenizer.eos_token_id,
        )

    def _tokenizer(self, model, **kwargs):
        # intantiate the tokenizer
        if isinstance(model, str):
            # make sure transformers is installed
            try:
                import transformers
            except:
                raise Exception(
                    "Please install transformers with `pip install transformers` in order to use guidance.models.togetherai!"
                )

            try:
                tokenizer = transformers.AutoTokenizer.from_pretrained(
                    model, use_fast=False, **kwargs
                )
                # This is here because some tokenizers are bad and don't have all the bytes (I'm looking at you, microsoft/phi2)
                if hasattr(tokenizer, "byte_decoder"):
                    all_bytes = set()
                    for x in tokenizer.get_vocab().keys():
                        [all_bytes.add(y) for y in x]
                    assert (
                        set(tokenizer.byte_decoder.keys()).intersection(all_bytes)
                        == all_bytes
                    )
            except:
                tokenizer = transformers.AutoTokenizer.from_pretrained(
                    model, use_fast=True, **kwargs
                )  # fall back to the fast tokenizer

        assert (
            tokenizer is not None
        ), "You must give a model name when you provide a tokenizer object!"

        return tokenizer

    def __call__(self, byte_string):
        tokenisation = self._orig_tokenizer(byte_string)
        return tokenisation["input_ids"]


class TransformersEngine(Engine):
    def __init__(self, model, tokenizer, compute_log_probs, **kwargs):
        # fill in default model value
        if model is None:
            model = os.environ.get("TRANSFORMERS_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser("~/.transformers_model"), "r") as file:
                    model = file.read().replace("\n", "")
            except:
                pass

        self.model_obj = self._model(model, **kwargs)

        if not isinstance(model, str):
            self.model = model.__class__.__name__
        self.device = self.model_obj.device  # otherwise note the current device

        self._past_key_values = None
        self._cached_logits = None
        self._cached_token_ids = []

        super().__init__(
            TransformersTokenizer(model, tokenizer), compute_log_probs=compute_log_probs
        )

    def _model(self, model, **kwargs):
        # intantiate the model if needed
        if isinstance(model, str):

            # make sure transformers is installed
            try:
                import transformers
            except:
                raise Exception(
                    "Please install transformers with `pip install transformers` in order to use guidance.models.Transformers!"
                )
            model = transformers.AutoModelForCausalLM.from_pretrained(model, **kwargs)
        return model

    def _joint_tokenize(self, token_ids):
        # first_decode = self.tokenizer._orig_tokenizer.decode(token_ids)

        # the encode/decode cycle might not work if we have partial unicode strings
        used_tokens = len(token_ids)
        for _ in range(3):
            try:
                first_decode = b"".join(
                    [self.tokenizer.tokens[id] for id in token_ids[:used_tokens]]
                ).decode("utf8")
            except UnicodeDecodeError:
                if used_tokens == 0:
                    break
                else:
                    used_tokens -= 1

        new_ids = self.tokenizer._orig_tokenizer(
            first_decode, add_special_tokens=False
        )["input_ids"]
        if used_tokens < len(token_ids):
            new_ids += token_ids[used_tokens:]

        # HACK: check for a bug in the HuggingFace tokenizer (that will just add extra spaces during an encode-decode cycle)
        second_decode = self.tokenizer._orig_tokenizer.decode(new_ids)
        if (
            second_decode != first_decode
            and len(second_decode) == len(first_decode) + 1
            and second_decode.startswith("<s>  ")
        ):
            new_ids = new_ids[0:1] + new_ids[2:]

        return new_ids

    def get_logits(self, token_ids, forced_bytes, current_temp):
        """Computes the logits for the given token state.

        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        """

        # make sure we don't run off the end of the model
        if len(token_ids) >= getattr(
            self.model_obj.config, "max_position_embeddings", 1e10
        ):
            raise Exception(
                f"Attempted to run a transformers model past its maximum context window size of {self.model_obj.config.max_position_embeddings}!"
            )

        # get the number of cache positions we are using
        cache_token_ids = self._cached_token_ids
        num_cached = 0
        for id in cache_token_ids:
            if (
                num_cached >= len(cache_token_ids)
                or num_cached >= len(token_ids)
                or token_ids[num_cached] != id
            ):
                break
            num_cached += 1

        # reset the cache length according to that number of positions
        past_key_values = self._past_key_values
        past_length = (
            past_key_values[0][0].size(-2) if past_key_values is not None else 0
        )
        if past_length > num_cached:
            past_length = max(
                0, num_cached - 1
            )  # note we recompute the last token because we don't bother to handle the special case of just computing logits
            self._past_key_values = tuple(
                tuple(p[..., :past_length, :] for p in v) for v in past_key_values
            )
        cache_token_ids[past_length:] = []

        # call the model
        new_token_ids = token_ids[past_length:]
        if len(new_token_ids) > 0:
            with torch.no_grad():
                model_out = self.model_obj(
                    input_ids=torch.tensor(new_token_ids).unsqueeze(0).to(self.device),
                    past_key_values=self._past_key_values,
                    use_cache=True,
                    position_ids=torch.arange(
                        past_length, past_length + len(new_token_ids)
                    )
                    .unsqueeze(0)
                    .to(self.device),
                    attention_mask=torch.ones(1, past_length + len(new_token_ids)).to(
                        self.device
                    ),
                    return_dict=True,
                    output_attentions=False,
                    output_hidden_states=False,
                )

            # save the results
            self._past_key_values = model_out.past_key_values
            cache_token_ids.extend(new_token_ids)
            # Need to add special truncating logic here for weird models that have a different output size than tokenizer vocab
            self._cached_logits = (
                model_out.logits[0, -1, : len(self.tokenizer.tokens)].cpu().numpy()
            )
            self.metrics.engine_input_tokens += len(new_token_ids)
            self.metrics.engine_output_tokens += 1

        return self._cached_logits


class Transformers(Model):
    def __init__(
        self, model=None, tokenizer=None, echo=True, compute_log_probs=False, **kwargs
    ):
        """Build a new Transformers model object that represents a model in a given state."""
        super().__init__(
            TransformersEngine(model, tokenizer, compute_log_probs, **kwargs), echo=echo
        )


class TransformersChat(Transformers, Chat):
    pass



import requests
import os
import base64

from ._model import Engine, EngineCallResponse


class RemoteEngine(Engine):
    """This connects to a remote guidance server and runs all computation using the remote engine."""

    def __init__(self, server_url, api_key, verify=None):
        self.server_url = server_url
        self.api_key = api_key
        if verify is None:
            verify = os.getenv("GUIDANCE_SSL_CERTFILE", None)
        self.verify_crt = verify

    def __call__(self, parser, grammar, ensure_bos_token=True):
        # Prepare the request data
        data = {
            "parser": parser,
            "grammar": base64.b64encode(grammar.serialize()).decode("utf-8"),
        }

        headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}

        # Send the request to the server
        response = requests.post(
            self.server_url + "/extend",
            json=data,
            headers=headers,
            stream=True,
            verify=self.verify_crt,
        )

        # Check for valid response
        if response.status_code != 200:
            response.raise_for_status()

        # Process and yield the response data
        for chunk in response.iter_content(
            chunk_size=None
        ):  # chunk_size=None means it'll stream the content
            response_data = EngineCallResponse.deserialize(chunk)
            yield response_data



import os
import pathlib
from typing import Any, Dict, Optional
from urllib.parse import urlparse, parse_qs

import papermill as pm
import pytest

BASE_NB_PATH = pathlib.Path("./notebooks").absolute()


def run_notebook(notebook_path: pathlib.Path, params: Optional[Dict[str, Any]] = None):
    assert notebook_path.exists(), f"Checking for: {notebook_path}"
    output_nb = notebook_path.stem + ".papermill_out" + notebook_path.suffix
    output_path = TestTutorials.BASE_TUTORIAL_PATH / output_nb

    # Just make sure nothing throws an exception
    pm.execute_notebook(
        input_path=notebook_path, output_path=output_path, parameters=params
    )


class TestTutorials:
    BASE_TUTORIAL_PATH = BASE_NB_PATH / "tutorials"

    @pytest.mark.needs_credentials
    def test_chat(self, rate_limiter):
        azureai_endpoint = os.getenv("AZUREAI_CHAT_ENDPOINT", None)

        parsed_url = urlparse(azureai_endpoint)
        parsed_query = parse_qs(parsed_url.query)
        azureai_deployment = pathlib.Path(parsed_url.path).parts[3]
        version = parsed_query["api-version"]
        min_azureai_endpoint = f"{parsed_url.scheme}://{parsed_url.netloc}"

        os.environ["AZUREAI_CHAT_BASE_ENDPOINT"] = min_azureai_endpoint
        os.environ["AZUREAI_CHAT_API_VERSION"] = version[0]
        os.environ["AZUREAI_CHAT_DEPLOYMENT"] = azureai_deployment
        nb_path = TestTutorials.BASE_TUTORIAL_PATH / "chat.ipynb"
        run_notebook(nb_path, params=dict(call_delay_secs=rate_limiter))

    def test_regex_constraints(self):
        nb_path = TestTutorials.BASE_TUTORIAL_PATH / "regex_constraints.ipynb"
        run_notebook(nb_path)

    def test_token_healing(self):
        nb_path = TestTutorials.BASE_TUTORIAL_PATH / "token_healing.ipynb"
        run_notebook(nb_path)


class TestModels:
    BASE_MODEL_PATH = BASE_NB_PATH / "api_examples" / "models"

    @pytest.mark.needs_credentials
    def test_azure_openai(self, rate_limiter):
        azureai_endpoint = os.getenv("AZUREAI_CHAT_ENDPOINT", None)

        parsed_url = urlparse(azureai_endpoint)
        parsed_query = parse_qs(parsed_url.query)
        azureai_deployment = pathlib.Path(parsed_url.path).parts[3]
        version = parsed_query["api-version"][0]
        min_azureai_endpoint = f"{parsed_url.scheme}://{parsed_url.netloc}"

        os.environ["AZUREAI_CHAT_BASE_ENDPOINT"] = min_azureai_endpoint
        os.environ["AZUREAI_CHAT_API_VERSION"] = version
        os.environ["AZUREAI_CHAT_DEPLOYMENT"] = azureai_deployment

        nb_path = TestModels.BASE_MODEL_PATH / "AzureOpenAI.ipynb"
        run_notebook(nb_path, params=dict(call_delay_secs=rate_limiter))


class TestArtOfPromptDesign:
    BASE_APD_PATH = BASE_NB_PATH / "art_of_prompt_design"

    @pytest.mark.skip(reason="Having trouble running")
    @pytest.mark.use_gpu
    def test_prompt_boundaries_and_token_healing(self):
        nb_path = (
            TestArtOfPromptDesign.BASE_APD_PATH
            / "prompt_boundaries_and_token_healing.ipynb"
        )
        run_notebook(nb_path)

    @pytest.mark.use_gpu
    def test_react(self, selected_model_name):
        if selected_model_name in ["phi2gpu"]:
            # I don't know why; it doesn't make sense, but
            msg = (
                f"react notebook disagrees with {selected_model_name}; reasons obscure"
            )
            pytest.skip(msg)
        nb_path = TestArtOfPromptDesign.BASE_APD_PATH / "react.ipynb"
        run_notebook(nb_path)

    @pytest.mark.use_gpu
    @pytest.mark.needs_credentials
    def test_use_clear_syntax(self, rate_limiter):
        azureai_endpoint = os.getenv("AZUREAI_CHAT_ENDPOINT", None)

        parsed_url = urlparse(azureai_endpoint)
        parsed_query = parse_qs(parsed_url.query)
        azureai_deployment = pathlib.Path(parsed_url.path).parts[3]
        version = parsed_query["api-version"]
        min_azureai_endpoint = f"{parsed_url.scheme}://{parsed_url.netloc}"

        os.environ["AZUREAI_CHAT_BASE_ENDPOINT"] = min_azureai_endpoint
        os.environ["AZUREAI_CHAT_API_VERSION"] = version[0]
        os.environ["AZUREAI_CHAT_DEPLOYMENT"] = azureai_deployment
        nb_path = TestArtOfPromptDesign.BASE_APD_PATH / "use_clear_syntax.ipynb"
        run_notebook(nb_path, params=dict(call_delay_secs=rate_limiter))



import pathlib

from typing import Type
from urllib.parse import parse_qs, urlparse

import tiktoken

from ._grammarless import Grammarless
from ._model import Chat, Instruct
from ._openai import (
    OpenAIChatEngine,
    OpenAICompletionEngine,
    OpenAIInstructEngine
)

try:
    import openai as openai_package

    is_openai = True
except ModuleNotFoundError:
    is_openai = False


class AzureOpenAI(Grammarless):
    """Represents an Azure OpenAI model as exposed through their remote API.

    Note that because this uses a remote API endpoint without built-in guidance support
    there are some things we cannot do, like force the model to follow a pattern inside
    a chat role block.

    Authentication can be provided via an `api_key` or through Entra/Azure Active
    Directory.
    """

    def __init__(
        self,
        model: str,
        azure_endpoint: str,
        azure_deployment: str = None,
        azure_ad_token_provider=None,
        api_key: str = None,
        tokenizer=None,
        echo=True,
        version: str = None,
        max_streaming_tokens=1000,
        timeout=0.5,
        compute_log_probs=False,
        **kwargs,
    ):
        """Build a new AzureOpenAI model object that represents a model in a given state.

        Parameters
        ----------
        model : str
            The name of the OpenAI model to use (e.g. gpt-3.5-turbo).
        azure_endpoint: str
            The endpoint of the deployed model (e.g. https://my_azureai_instance.openai.azure.com)
        azure_deployment: str
            The deployed name of the model (given when the deployment was created)
        api_key: str
            The API key for calling the model
        azure_ad_token_provider:
            Alternative to the api_key, allows for use of Azure Entra authentication
        """
        if not is_openai or not hasattr(openai_package, "OpenAI"):
            raise ImportError(
                "Please install the openai package version >= 1 using `pip install openai -U` "
                "in order to use guidance.models.OpenAI!"
            )

        if api_key is None and azure_ad_token_provider is None:
            raise ValueError("Please provide either api_key or azure_ad_token_provider")

        # if we are called directly (as opposed to through super()) then we convert ourselves to
        # a more specific subclass if possible
        if self.__class__ is AzureOpenAI:
            # Default to a completion model
            found_subclass: Type[AzureOpenAI] = (
                AzureOpenAICompletion
                if model.endswith("-instruct") 
                else AzureOpenAIChat
            )

            # convert to any found subclass
            self.__class__ = found_subclass
            found_subclass.__init__(
                self,
                model=model,
                azure_endpoint=azure_endpoint,
                api_key=api_key,
                azure_ad_token_provider=azure_ad_token_provider,
                azure_deployment=azure_deployment,
                tokenizer=tokenizer,
                echo=echo,
                version=version,
                **kwargs,
            )
            return
        
        parsed_url = urlparse(azure_endpoint)

        if azure_deployment is None:
            parts = pathlib.Path(parsed_url.path).parts
            if len(parts) > 2:
                azure_deployment = parts[3]
                
        parsed_query = parse_qs(parsed_url.query)
        api_version = (
            version
            if "api-version" not in parsed_query
            else parsed_query["api-version"]
        )
        engine_map = {
            AzureOpenAICompletion: OpenAICompletionEngine,
            AzureOpenAIChat: OpenAIChatEngine,
            AzureOpenAIInstruct: OpenAIInstructEngine,
        }
        engine_class = engine_map[self.__class__]

        if tokenizer is None:
            tokenizer = tiktoken.encoding_for_model(model)

        engine_instance = engine_class(
            tokenizer=tokenizer,
            max_streaming_tokens=max_streaming_tokens,
            timeout=timeout,
            compute_log_probs=compute_log_probs,
            model=model,
            azure_endpoint=azure_endpoint,
            api_key=api_key,
            azure_ad_token_provider=azure_ad_token_provider,
            api_version=api_version,
            azure_deployment=azure_deployment,
            client_class=openai_package.AzureOpenAI,
            **kwargs,
        )

        super().__init__(
            engine_instance,
            echo=echo,
        )


class AzureOpenAIChat(AzureOpenAI, Chat):
    pass


class AzureOpenAICompletion(AzureOpenAI):
    pass


class AzureOpenAIInstruct(AzureOpenAI, Instruct):
    def get_role_start(self, name):
        return ""

    def get_role_end(self, name):
        if name == "instruction":
            return "<|endofprompt|>"
        else:
            raise ValueError(
                f"The OpenAIInstruct model does not know about the {name} role type!"
            )


import threading
import numpy as np
import queue
import time
import tiktoken
import re
import logging
from ._model import Tokenizer, Engine, Model, format_pattern, ConstraintException

logger = logging.getLogger(__name__)


class GrammarlessTokenizer(Tokenizer):
    def __init__(self, tokenizer):

        # Grammarless models don't always have public tokenizations, so when not provided we pretend they tokenize like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")

        # tiktoken tokenizer was given
        if hasattr(tokenizer, "decode_tokens_bytes"):
            special_map = {v: k for k, v in tokenizer._special_tokens.items()}
            first_special = tokenizer.n_vocab
            for k in special_map:
                if k < first_special:
                    first_special = k

            # consume one-by-one until we have passed all the special tokens AND gotten a valid token
            i = tokenizer.n_vocab - 1
            byte_tokens = []
            while True:
                try:
                    bval = tokenizer.decode_single_token_bytes(i)
                    found = True
                except KeyError:
                    bval = special_map.get(i, b"<|invalid_special_token|>")
                    found = False
                byte_tokens.append(bval)

                if i < first_special and found:
                    break
                i -= 1

            # do the rest of the tokens as a batch
            byte_tokens = tokenizer.decode_tokens_bytes(np.arange(i + 1)) + byte_tokens

            if hasattr(tokenizer, "bos_token_id"):
                bos_token_id = tokenizer.bos_token_id
            else:
                bos_token_id = None
            if hasattr(tokenizer, "eos_token_id"):
                eos_token_id = tokenizer.eos_token_id
            else:
                eos_token_id = tokenizer._special_tokens["<|endoftext|>"]

        # a transformer tokenizer was given that has a byte_decoder
        elif hasattr(tokenizer, "byte_decoder"):
            byte_tokens = []
            for i in range(tokenizer.vocab_size):
                byte_coded = bytes(
                    [
                        tokenizer.byte_decoder[c]
                        for c in tokenizer.convert_ids_to_tokens(i)
                    ]
                )
                byte_tokens.append(byte_coded)
            bos_token_id = tokenizer.bos_token_id
            eos_token_id = tokenizer.eos_token_id

        # a transformer tokenizer was given with byte_decoder
        elif hasattr(tokenizer, "convert_ids_to_tokens"):
            byte_tokens = [
                bytes(
                    tokenizer.convert_tokens_to_string(
                        ["a", tokenizer.convert_ids_to_tokens(i)]
                    )[1:],
                    encoding="utf8",
                )
                for i in range(tokenizer.vocab_size)
            ]
            bos_token_id = tokenizer.bos_token_id
            eos_token_id = tokenizer.eos_token_id

        # a HuggingFace tokenizers tokenizer was given with id_to_token
        elif hasattr(tokenizer, "id_to_token"):
            a_token_ids = tokenizer.encode("a").ids
            if len(a_token_ids) == 3:
                bos_token_id = a_token_ids[0]
                a_id = a_token_ids[1]
                eos_token_id = a_token_ids[2]
            else:
                raise Exception(
                    "This tokenizer does not seem to have a BOS and EOS, support for this need to be implemented still."
                )

            byte_tokens = [
                bytes(tokenizer.decode([a_id, i])[1:], encoding="utf8")
                for i in range(tokenizer.get_vocab_size())
            ]
            for i, b in enumerate(byte_tokens):
                if b == b"":
                    byte_tokens[i] = bytes(tokenizer.id_to_token(i), encoding="utf8")

        else:
            raise Exception("The tokenizer given was not of a recognized type!")

        super().__init__(byte_tokens, bos_token_id, eos_token_id)


class GrammarlessEngine(Engine):
    def __init__(self, tokenizer, max_streaming_tokens, timeout, compute_log_probs):
        self.max_streaming_tokens = max_streaming_tokens
        self.timeout = timeout

        self._data_queue = (
            queue.Queue()
        )  # this is where the streaming thread puts results
        self._data = b""  # these are the bytes we are ready to use in the main thread
        self._not_running_stream = (
            threading.Event()
        )  # this is phrased negatively so we can wait for the stop event
        self._last_call = 0
        self._num_calls_made = 0
        self._current_temp = 0
        self._last_stream_start = None

        self._not_running_stream.set()

        self.max_repeated_calls = 10
        self.timeout = timeout

        # If tokenizer is not already an instance of Tokenizer, then instantiate it as a GrammarlessTokenizer
        if not isinstance(tokenizer, Tokenizer):
            tokenizer = GrammarlessTokenizer(tokenizer)

        # build the
        super().__init__(tokenizer=tokenizer, compute_log_probs=compute_log_probs)

    def __call__(self, *args, **kwargs):
        self._num_calls_made = 0  # reset the number of calls count so we only limit the number of calls within a single grammar execution
        return super().__call__(*args, **kwargs)

    def _running_stream(self):
        return not self._not_running_stream.is_set()  # wrap double negation (which)

    def _start_generator_stream(self, generator):
        logger.debug(f"start Grammarless._start_generator_stream")
        dqueue = self._data_queue
        first_iteration = True
        try:
            for chunk in generator:
                logger.debug(f"Got chunk: " + str(chunk))
                if len(chunk) > 0:
                    dqueue.put(chunk)
                if (
                    self._not_running_stream.is_set()
                    or not first_iteration
                    and time.time() - self._last_call > self.timeout
                ):
                    break
                first_iteration = False

        # we pass any exceptions back to the main thread
        except Exception as e:
            self._not_running_stream.set()
            while not dqueue.empty():
                dqueue.get()
            dqueue.put(e)

        if self._running_stream():
            dqueue.put(self.tokenizer.eos_token)
        self._not_running_stream.set()
        dqueue.put(
            b""
        )  # so we never get stuck waiting for a running stream to return something

    def _start_new_stream(self, prompt, temperature):

        # make sure the display is up to date (since we are about to delay for a while)
        # TODO: how can we handle this better since the engine is now separate from the client?
        #       we could use a timeout for the GUI update throttling, those were just kind of slow... (but would be best)
        # self._update_display(throttle=False)

        if self._num_calls_made > self.max_repeated_calls:
            raise Exception(
                f"We have exceeded the maximum number of repeat calls ({self.max_repeated_calls}) per grammar execution!"
            )

        # stop any running stream
        if self._running_stream():
            self._not_running_stream.set()  # stop the generator
            self._remote_thread.join()  # wait for the thread to finish

        # clear the data queue
        while not self._data_queue.empty():
            self._data_queue.get()

        # start the new stream
        self._used_bytes_len = 0
        self._current_temp = temperature
        self._last_call = time.time()
        generator = self._generator(prompt, temperature)
        self._not_running_stream.clear()  # so we know we are running
        self._num_calls_made += 1
        self._remote_thread = threading.Thread(
            target=self._start_generator_stream, args=(generator,)
        )
        self._remote_thread.start()

    def _reset_shared_data(self, new_data, temperature):
        """Should be called by _generator calls to reset the shared data state."""
        if temperature == 0 and self._last_stream_start == new_data:
            raise self._report_failed_match(new_data)
        self._data = new_data
        self._last_stream_start = self._data

    def get_logits(self, token_ids, forced_bytes, current_temp):
        """Computes the logits for the given token state.

        This overrides a method from the Local class that is used to get
        inference results from the model.
        """

        logger.debug(f"start Grammarless._get_logits(token_ids={token_ids})")

        if len(token_ids) == 0:
            raise ValueError("token_ids must contain some tokens.")

        # compute the prompt bytes
        whole_token_prompt = b"".join([self.tokenizer.tokens[i] for i in token_ids])
        prompt = whole_token_prompt + forced_bytes

        self._last_call = time.time()

        # keep looping until we have at least one more byte past our prompt
        token_id = None
        restarted = False  # track if we have restarted the data stream during this call
        while True:

            # if the generation temperature changes we have to restart
            if self._current_temp != current_temp:
                self._start_new_stream(prompt, current_temp)
                continue

            # try and get the next token id
            elif self._data.startswith(prompt):
                token_id = self._get_next_token(len(prompt) - len(forced_bytes))
                if token_id is not None:

                    # if we have a non-zero sampling temperature we can't reuse bytes
                    new_used_len = len(whole_token_prompt) + len(
                        self.tokenizer.tokens[token_id]
                    )
                    if current_temp > 0 and self._used_bytes_len >= new_used_len:
                        token_id = None
                        self._start_new_stream(prompt, current_temp)
                        continue

                    # ...otherwise we have found the token id we want to emit
                    else:
                        self._used_bytes_len = len(whole_token_prompt) + len(
                            self.tokenizer.tokens[token_id]
                        )
                        break

            # restart if extending our data will never lead to matching our prompt
            elif not self._data.startswith(prompt) and len(self._data) >= len(
                prompt
            ):  # not prompt.startswith(self._data): # len(self._data) >= len(prompt) or

                # check if we have already restarted once and so retrying by default is not likely to be helpful
                if restarted:
                    raise self._report_failed_match(prompt)

                # check the length of the prefix match
                match_len = 0
                found_mismatch = False
                data = self._data
                for match_len, v in enumerate(prompt):
                    if v != data[match_len]:
                        found_mismatch = True
                        break
                if not found_mismatch:
                    match_len = len(prompt)
                leftover = prompt[match_len:]

                # record any active non-empty role ends. Ignore role ends that are spaces
                parts = [
                    b"<|im_end|>",
                    self.tokenizer.eos_token,
                ]  # note we assume we are role tags that end with <|im_end|>

                # for _,role_end_str in self.opened_blocks.values():
                #     role_end_str = format_pattern.sub("", role_end_str)
                #     if len(role_end_str) > 0 and not re.fullmatch(r'\s+', role_end_str):
                #         parts.append(role_end_str.encode("utf8"))

                # # record the eos token
                # parts.append(self.eos_token)

                # see if adding an end token would work here (if so we avoid recalling the server and just produce an end token)
                found_match = False
                for p in parts:
                    if p.startswith(leftover):
                        self._data = self._data[:match_len] + p
                        logger.debug(
                            f"automatically adding an end token since it fits the forcing of the grammar"
                        )
                        found_match = True
                        break
                if found_match:
                    continue  # start our loop over again

                logger.debug(
                    f"restarting a stream because the data we have does not match the ids. We have {str(self._data)} but the prompt is {str(prompt)}"
                )
                restarted = True
                self._start_new_stream(prompt, current_temp)

            # extend our data with a chunk from the model stream
            if not self._data_queue.empty():
                new_bytes = self._data_queue.get_nowait()
                if isinstance(new_bytes, Exception):
                    raise new_bytes

                # if we are at the end of the generation then we try again allowing for early token stopping
                if len(new_bytes) == 0:
                    token_id = self._get_next_token(len(prompt), allow_early_stop=True)
                    if token_id is not None:
                        break
                self._data += new_bytes

            # but if there is nothing and we are not running then we start a stream
            elif self._not_running_stream.is_set():
                logger.debug(
                    "starting a new stream because there is no data to read and no stream running..."
                )
                restarted = True
                self._start_new_stream(prompt, current_temp)

            # we wait for the running stream to put something in the queue
            else:
                self._last_call = 10e9  # set to essentialy infinity so we don't stop the data stream while we are waiting for it
                new_bytes = self._data_queue.get()
                if isinstance(new_bytes, Exception):
                    raise new_bytes
                self._data += new_bytes
                self._last_call = (
                    time.time()
                )  # reset out call time to allow the data stream to time out if we happen to be done with it

        # # if we don't have the next byte of data yet then we wait for it (from the streaming thread)
        # if len(self._data) == len(prompt):
        #     self._data += self._data_queue.get()

        # token_id = self._get_next_token(len(prompt))

        # set the logits to the next byte the model picked
        logits = np.ones(len(self.tokenizer.tokens)) * -np.inf
        logits[token_id] = 100
        if token_id != self.tokenizer.eos_token:
            logits[self.tokenizer.eos_token_id] = (
                0  # we always allow the model to use EOS if that is the only way forward
            )

        return logits

    def _report_failed_match(self, prompt):

        # check the length of the prefix match
        match_len = 0
        found_mismatch = False
        data = self._data
        for match_len, v in enumerate(prompt):
            if v != data[match_len]:
                found_mismatch = True
                break
        if not found_mismatch:
            match_len = len(prompt)
        leftover = prompt[match_len:]

        # compute the mismatch parts
        data_after_prompt = self._data[match_len:]
        if len(data_after_prompt) > 40:
            data_after_prompt = data_after_prompt[:40] + b"..."
        prompt_tail = prompt[:match_len]
        if len(prompt_tail) > 40:
            prompt_tail = b"..." + prompt_tail[-40:]

        # show in the model output where and how we diverged from the grammar
        try:
            # just for display when echo is on
            already_shown = len(self._current_prompt().encode())
            self += (
                self._data[already_shown:match_len].decode()
                + f"<||_html:<span style='color: rgba(165,0,0,1);' title='{leftover}'><span style='text-decoration: underline;'>{data_after_prompt.decode()}</span></span>_||>"
            )
        except:
            pass  # could not decode the data the model generated into a string...

        # create an exception for users to deal with (that our caller can throw)
        return ConstraintException(
            f"The model attempted to generate {str(data_after_prompt)} after the prompt `{prompt_tail}`, but that does\n"
            + "not match the given grammar constraints! Since your model is a remote API that does not support full guidance\n"
            + "integration we cannot force the model to follow the grammar, only flag an error when it fails to match.\n"
            + "You can try to address this by improving the prompt, making your grammar more flexible, rerunning with\n"
            + "a non-zero temperature, or using a model that supports full guidance grammar constraints."
        )

    def _get_next_token(self, pos, allow_early_stop=False):
        data = self._data
        trie = self._token_trie
        token_id = None
        while True:

            # see if we have run out of data
            if pos >= len(data):
                if allow_early_stop:
                    return token_id
                else:
                    return None

            # try and walk down the trie
            next_byte = data[pos : pos + 1]
            if trie.has_child(next_byte):
                trie = trie.child(next_byte)
                pos += 1
                if trie.value >= 0:
                    token_id = trie.value
            else:
                return token_id  # this is the longest greedy token match we can make


class Grammarless(Model):
    """The base class for all remote models (hosted behind a remote API)."""

    pass

    # def __init__(self, model, tokenizer=None, echo=True, compute_log_probs=False, max_streaming_tokens=None, timeout=0.5):
    #     '''Build a new remote model object that represents a model in a given state.

    #     This is an abstract class. To instantiate it use a specific subclass like guidance.models.OpenAI.
    #     '''
    #     super().__init__(
    #         engine=GrammarlessEngine(model, tokenizer, max_streaming_tokens, timeout, compute_log_probs),
    #         echo=echo
    #     )



from sys import stderr
import numpy as np
from ordered_set import OrderedSet
from ._grammar import Join, Select, Terminal, Null, Byte, ByteRange


class ParserException(Exception):
    def __init__(self, *args, **kwargs):
        self.current_byte = kwargs.pop("current_byte", None)
        self.allowed_bytes = kwargs.pop("allowed_bytes", None)
        super().__init__(*args, **kwargs)


class EarleyItem:
    __slots__ = (
        "node",
        "values",
        "start",
        "pos",
        "log_prob",
        "children",
        "hidden_start",
    )

    def __init__(self, node, values, pos, start, log_prob, hidden_start):
        self.node = node
        self.values = values
        self.start = start
        self.pos = pos
        self.log_prob = log_prob
        self.children = None
        self.hidden_start = hidden_start

    def __eq__(self, other):
        return (
            isinstance(other, EarleyItem)
            and self.start == other.start
            and self.pos == other.pos
            and self.node == other.node
            and self.values == other.values
            and self.log_prob == other.log_prob
        )

    def __hash__(self):
        return hash((self.node, self.values, self.start, self.pos))

    def __repr__(self):
        if isinstance(self.node, Join):
            s = f"{self.node.name:20} -> "
            rs = ""
            for i, v in enumerate(self.values):
                if self.pos == i:
                    rs += "•"
                rs += v.name + " "
            if self.pos == len(self.values):
                rs += "•"
        elif isinstance(self.node, Select):
            s = f"{self.node.name:20} -> "
            rs = ""
            if self.pos == 0:
                rs += "•"
            rs += self.values[0].name
            if self.pos == 1:
                rs += "•"
        else:
            assert False
        return s + f"{rs:40} ({self.start}) {'nullable' if self.node.nullable else ''}"


class Parser:
    """An abstract base class for guidance parsers."""

    pass


class EarleyCommitParser(Parser):
    def __init__(self, grammar):

        # we can't have a terminal as the root
        if isinstance(grammar, Terminal):
            grammar = Join([grammar])

        self.grammar = grammar
        self.bytes = b""
        self.state_sets = [OrderedSet()]  # the list of Earley items for each byte
        self.token_counts = []  # used to track how many tokens have been used
        self.state_set_pos = 0
        self.shadow_pos = 0
        self._add_node(self.grammar, 0, 0.0, 1000000000)
        self._inner_loop(self.state_set_pos)

    @property
    def pos(self):
        return self.shadow_pos

    @pos.setter
    def pos(self, new_pos):

        # do nothing if we aren't moving
        if new_pos == self.state_set_pos:
            return
        elif new_pos > self.state_set_pos:
            raise ParserException(
                "Can't move the parser position forward! (only backward)"
            )

        # check if we are just moving the shadow position
        if new_pos >= self.shadow_pos:
            self.shadow_pos = new_pos
            return

        # actually reset our position if we need to
        self.state_sets = self.state_sets[: new_pos + 1] + [OrderedSet()]
        self.token_counts = self.token_counts[: new_pos + 2]
        self.bytes = self.bytes[:new_pos]
        self.state_set_pos = new_pos
        self.shadow_pos = new_pos
        self._inner_loop(self.state_set_pos)

    def _add_item(self, state_set_pos, new_item):
        state_set = self.state_sets[state_set_pos]
        if new_item not in state_set:
            state_set.append(new_item)
        else:
            existing_item = state_set.items[state_set.map[new_item]]
            existing_item.hidden_start = min(
                existing_item.hidden_start, new_item.hidden_start
            )

    def _add_node(self, grammar, state_set_pos, log_prob, hidden_start):
        if isinstance(grammar, Terminal):
            new_item = EarleyItem(
                grammar, tuple(), 0, state_set_pos, log_prob, hidden_start
            )
            self._add_item(state_set_pos, new_item)

        elif isinstance(grammar, Join):
            new_item = EarleyItem(
                grammar, tuple(grammar.values), 0, state_set_pos, log_prob, hidden_start
            )
            self._add_item(state_set_pos, new_item)

        elif isinstance(grammar, Select):
            for value in grammar.values:
                new_item = EarleyItem(
                    grammar, (value,), 0, state_set_pos, log_prob, hidden_start
                )
                self._add_item(state_set_pos, new_item)

    def _inner_loop(self, state_set_pos, start_pos=0):
        curr_state_set = self.state_sets[state_set_pos]
        if len(self.state_sets) == state_set_pos + 1:
            self.state_sets.append(OrderedSet())
            self.token_counts.append(
                self.token_counts[-1] if len(self.token_counts) > 0 else 0
            )
        next_state_set = self.state_sets[state_set_pos + 1]
        pos = start_pos
        while len(curr_state_set) > pos:
            item = curr_state_set[pos]

            # completion
            if item.pos == len(item.values):

                # if we complete an item that is a "commit point" then we eliminate all other possible
                # parses so that we are "committed" to using this item
                # we do this by removing any unprocessed items in the current state set and clearing the next state set
                if item.node.commit_point:
                    while len(curr_state_set) > pos:

                        # if we find another valid commit point that starts earlier we use that instead
                        # this causes us to pick the longest matching valid commit point
                        end_item = curr_state_set[-1]
                        if (
                            end_item.node.commit_point
                            and end_item.pos == len(end_item.values)
                            and end_item.start < item.start
                        ):
                            item = end_item

                        curr_state_set.pop()
                    curr_state_set.append(
                        item
                    )  # we append the current item again (we do this since we may have swapped it out above)
                    next_state_set.clear()

                # advance all the parents that our completion impacts
                token_span = (
                    self.token_counts[state_set_pos] - self.token_counts[item.start]
                )
                start_state_set = self.state_sets[item.start]
                for start_item in start_state_set:
                    if (
                        start_item.pos < len(start_item.values)
                        and start_item.values[start_item.pos] == item.node
                    ):

                        # if item.node.max_tokens <= token_span and any(start_item.node == v and len(v.values) > 1 for v in item.node.values):
                        #     continue # skip advancing parents that are also children (recursion) once we are past the token limit

                        curr_state_set.append(
                            EarleyItem(
                                start_item.node,
                                start_item.values,
                                start_item.pos + 1,
                                start_item.start,
                                start_item.log_prob
                                + item.log_prob,  # increment the log prob by the child value,
                                start_item.hidden_start,
                            )
                        )

            # don't advance past our max token limit
            elif (
                item.node.max_tokens
                > self.token_counts[state_set_pos] - self.token_counts[item.start]
            ):

                # scan (note we only scan forward when we have more max token headroom left)
                next_item_node = item.values[item.pos]
                hidden_start = item.hidden_start
                if next_item_node.hidden:
                    hidden_start = min(state_set_pos, hidden_start)
                if isinstance(
                    next_item_node, Terminal
                ):  # and item.node.max_tokens > self.token_counts[state_set_pos] - self.token_counts[item.start]:
                    next_state_set.append(
                        EarleyItem(
                            item.node,
                            item.values,
                            item.pos + 1,
                            item.start,
                            item.log_prob,
                            hidden_start,
                        )
                    )  # the log prob will get incremented when consume_bytes is called

                # prediction
                else:
                    self._add_node(
                        next_item_node, state_set_pos, 0.0, hidden_start
                    )  # the log probs will get incremented by children later

                # handle nullable items by advancing them automatically (since we know we can)
                if next_item_node.nullable:
                    new_item = EarleyItem(
                        item.node,
                        item.values,
                        item.pos + 1,
                        item.start,
                        item.log_prob,
                        item.hidden_start,
                    )
                    if new_item not in self.state_sets[state_set_pos]:
                        self.state_sets[state_set_pos].append(new_item)
            pos += 1

    def earliest_hidden_start(self, state_pos=None):
        """The earliest that a hidden node might match.

        This is useful because it tells us which bytes may end being hidden.
        """
        if state_pos is None:
            state_pos = self.state_set_pos
        earliest_pos = 10000000000
        for item in self.state_sets[state_pos]:
            earliest_pos = min(earliest_pos, item.hidden_start)
        return earliest_pos

    def matched(self):
        """Checks if the parser has completely matched the grammar."""
        if self.shadow_pos != self.state_set_pos:
            return False
        for item in self.state_sets[self.state_set_pos]:
            if item.node == self.grammar and item.pos == len(item.values):
                return True
        return False

    def shadow_rewind(self, new_pos):
        if new_pos == self.state_set_pos:
            return
        self.shadow_pos = new_pos

    def commit_and_collapse_item(self, item):
        """This collapses the item into zero width and rewinds the parser position accordingly.

        Note we assume the item is in the current state set.
        """

        # trim off the state sets that matches this item
        self.state_sets = self.state_sets[: item.start + 1]
        self.token_counts = self.token_counts[: item.start + 1]
        self.bytes = self.bytes[: item.start]
        self.state_set_pos = item.start
        self.shadow_pos = item.start

        # add this state to its start point (making it a zero length match with no values)
        self.state_sets[item.start].append(
            EarleyItem(
                item.node, tuple(), 0, item.start, item.log_prob, item.hidden_start
            )
        )

        # expand from this state
        self._inner_loop(item.start, len(self.state_sets[item.start]) - 1)

    def mark_new_token(self):
        # TODO: we allow ourselves to go one past our max token limit when we hit a one-byte token
        #       because we don't know if we are continuing or extending a new token when we parse
        #       the first byte of the token. We could fix this by rerunning the inner_loop after each
        #       token, but we skip that for now since max_tokens is not a hard garuntee anyway when you
        #       have patterns.

        self.token_counts[-1] += 1

    def consume_byte(self, byte, log_prob=0.0):
        """Advances the parser by the given byte."""

        # see if we need to advance our shadow position...
        if self.shadow_pos < self.state_set_pos:
            assert (
                byte == self.bytes[self.shadow_pos : self.shadow_pos + 1]
            ), "Attempted to consume a byte by advancing shadow_pos but the byte didn't match!"
            self.shadow_pos += 1
            return

        # ...if not, we extend our bytes
        self.bytes += byte

        # filter out all the extensions that don't match this byte
        new_next_state_set = []
        found_valid = False
        found_invalid = False
        hidden_start = 10000000000
        for item in self.state_sets[self.state_set_pos + 1]:
            token_span = self.token_counts[-1] - self.token_counts[item.start]
            if item.node.max_tokens <= token_span:
                found_invalid = True
                continue
            elif item.pos > 0 and isinstance(item.values[item.pos - 1], Terminal):
                last_inner_node = item.values[item.pos - 1]
                if not last_inner_node.match_byte(byte):
                    found_invalid = True
                    continue
                else:
                    found_valid = True
                    if last_inner_node.commit_point:
                        item.log_prob += log_prob
                        new_next_state_set = [item]
                        hidden_start = min(hidden_start, item.hidden_start)
                        found_invalid = True  # we make everything else invalid, so that means we found something invalid
                        break
            item.log_prob += log_prob  # update the probability of the item by the probability of choosing this byte
            new_next_state_set.append(item)
            hidden_start = min(hidden_start, item.hidden_start)
        if not found_valid:
            raise ParserException(
                "Attempted to consume a byte that the grammar does not accept!",
                current_byte=byte,
            )
        if found_invalid:  # only update if we changed the set
            self.state_sets[self.state_set_pos + 1] = OrderedSet(new_next_state_set)

        # advance the parser one position
        self.state_set_pos += 1
        self.shadow_pos += 1
        self._inner_loop(self.state_set_pos)

        # look for a commit point node
        commit_point = None
        for item in self.state_sets[self.state_set_pos]:
            if (
                item.node.commit_point
                and item.pos == len(item.values)
                or (item.pos > 0 and item.values[item.pos - 1].commit_point)
            ):
                commit_point = item
                break  # TODO: consider how we might need to prioritize multiple commit point nodes (an uncommon scenario I think)
        # hidden_start,
        return commit_point

    def valid_next_bytes(self):
        """A list of Byte and ByteRange objects representing the next valid bytes."""
        valid_items = set()
        next_state_set = self.state_sets[self.state_set_pos + 1]
        for item in next_state_set:
            token_span = self.token_counts[-1] - self.token_counts[item.start]
            if item.node.max_tokens <= token_span:
                continue
            elif item.pos > 0 and isinstance(item.values[item.pos - 1], Terminal):
                v = item.values[item.pos - 1]
                if v not in valid_items:
                    valid_items.add(v)
        return valid_items

    def next_byte_temperature(self):
        """The maximum temperature over all the next bytes, or -1 if no temperature is set."""
        max_temp = -1
        next_state_set = self.state_sets[self.state_set_pos + 1]
        for item in next_state_set:
            if item.pos > 0 and isinstance(item.values[item.pos - 1], Terminal):
                v = item.values[item.pos - 1]
                max_temp = max(max_temp, v.temperature)
        return max_temp

    def next_byte_mask(self):
        """A mask version of the `valid_next_bytes` method."""

        mask = np.zeros(256, dtype=bool)

        # if we are shadow rewound then we just force those bytes again
        if self.shadow_pos < self.state_set_pos:
            mask[self.bytes[self.shadow_pos]] = True

        # otherwise we compute the valid bytes from the grammar
        else:
            valid_items = self.valid_next_bytes()
            for item in valid_items:
                if isinstance(item, Byte):
                    mask[item.byte[0]] = True
                elif isinstance(item, ByteRange):
                    mask[item.byte_range[0] : item.byte_range[1] + 1] = True
                else:
                    raise ParserException(
                        "Unknown Terminal Type: " + str(type(item)),
                    )
        return mask

    def __repr__(self, state_sets=None) -> str:
        s = ""
        if state_sets is None:
            _state_sets = self.state_sets
        else:
            _state_sets = state_sets
        for i, states in enumerate(_state_sets):
            s += f"\n=== {i} ==="
            if self.state_set_pos == i:
                s += " (state_set_pos)"
            s += "\n"
            for state in states:
                if isinstance(state.node, Join):
                    s += f"{state.node.name:20} -> "
                    rs = ""
                    for i, v in enumerate(state.values):
                        if state.pos == i:
                            rs += "•"
                        rs += v.name + " "
                    if state.pos == len(state.values):
                        rs += "•"
                elif isinstance(state.node, Select):
                    s += f"{state.node.name:20} -> "
                    rs = ""
                    if state.pos == 0:
                        rs += "•"
                    if len(state.values) == 0:
                        rs += "NO_VALUES!"
                    else:
                        rs += state.values[0].name
                        if state.pos == 1:
                            rs += "•"
                else:
                    assert False
                s += f"{rs:40} ({state.start}) {'nullable' if state.node.nullable else ''}\n"  # type: ignore[attr-defined]
        return s

    def _reversed_state_sets(self):
        new_state_sets = [OrderedSet([]) for _ in range(len(self.state_sets))]
        for i, states in enumerate(self.state_sets):
            for state in states:
                # if state.node.name == "__call___c":
                #     pass
                new_state_sets[state.start].append(
                    EarleyItem(
                        state.node,
                        state.values,
                        state.pos,
                        i,
                        state.log_prob,
                        state.hidden_start,
                    )
                )

        return new_state_sets

    def parse_tree(self):
        reversed_state_sets = self._reversed_state_sets()
        root_item = None

        # find the matching root state
        for item in reversed_state_sets[0]:
            if (
                item.node == self.grammar
                and item.start == len(self.bytes)
                and item.pos == len(item.values)
            ):  # note that ".start" mean end because items are reversed
                root_item = item
        if root_item is None:
            return None
        self._compute_parse_tree(0, root_item, reversed_state_sets)
        return root_item

    def get_captures(self, data=None, log_prob_data=None):
        root_node = self.parse_tree()
        if data is None:
            data = {}
        if log_prob_data is None:
            log_prob_data = {}
        if root_node is not None:
            # parse complete, so we can get the captures
            self._record_captures_from_root(root_node, data, log_prob_data)
            return data, log_prob_data
        # compute on partially parsed tree
        self._record_captures_partial(data, log_prob_data)
        return data, log_prob_data

    def _record_captures_partial(self, data, log_prob_data):
        byte_data = self.bytes

        for item in self.state_sets[self.state_set_pos]:
            cname = item.node.capture_name
            if cname is None:
                continue
            captured_value = byte_data[item.start : self.earliest_hidden_start()]
            if captured_value.endswith(b"<"):
                print(
                    "WARNING: Captured value ends with '<' which is a special character in the parser!",
                    file=stderr,
                )
            data[cname] = captured_value
            log_prob_data[cname] = item.log_prob

    def _record_captures_from_root(self, initial_item, data, log_prob_data):
        byte_data = self.bytes
        stack = [(initial_item, 0)]
        used_names = (
            set()
        )  # track which capture names have been used so self-recursive children don't overwrite their parents

        while stack:
            item, byte_pos = stack.pop()
            # terminal nodes
            if isinstance(item, Terminal):

                # if we are at a capture group node then we save the matched terminal byte
                if item.capture_name is not None:
                    data[item.capture_name] = item.byte
                    log_prob_data[item.capture_name] = 0

            # internal nodes
            else:
                start_byte_pos = byte_pos

                # recurse for all our non-null children
                for child in item.children:
                    if child is not None:
                        stack.append((child, byte_pos))
                        # _record_captures(child, data, log_prob_data, byte_data, byte_pos)
                        if isinstance(child, Terminal):
                            byte_pos += len(child)
                        else:
                            byte_pos = (
                                child.start
                            )  # note that "start" means "end" since this is a reversed state set

                # if we are at a capture group node then we save the matched bytes range
                # note that we record this after calling our children so that we save the outermost version of self-recursive calls
                cname = item.node.capture_name
                if (
                    cname is not None
                    and cname not in used_names
                    and not item.node.hidden
                ):

                    # see if we are doing a list append
                    if cname.startswith("__LIST_APPEND:"):
                        cname = cname[14:]  # trim off the list append tag
                        if cname not in data or not isinstance(data[cname], list):
                            data[cname] = []
                            log_prob_data[cname] = []
                        data[cname].append(byte_data[start_byte_pos : item.start])
                        log_prob_data[cname].append(item.log_prob)

                    # or just a regular assignment
                    else:
                        data[cname] = byte_data[
                            start_byte_pos : item.start
                        ]  # note that "start" means "end" since this is a reversed state set
                        log_prob_data[cname] = item.log_prob

                    used_names.add(cname)

    def _compute_parse_tree(self, initial_pos, initial_item, reversed_state_sets):
        stack = [(initial_pos, initial_item)]

        while stack:
            pos, item = stack.pop()

            # compute the children for this item
            assert self._compute_children(pos, item, reversed_state_sets)

            # recurse on the children
            for child in item.children:
                if child is None:
                    pass  # this child was nullable and was chosen to be null (empty)
                elif isinstance(child, Terminal):
                    pos += len(child)
                else:
                    stack.append((pos, child))
                    pos = (
                        child.start
                    )  # note that ".start" mean end because items are reversed

    def _compute_children(self, state_set_pos, item, reversed_state_sets, values_pos=0):

        # ensure we have a children array
        if item.children is None:
            item.children = [None for _ in range(len(item.values))]

        # consume as many terminal children as possible
        while True:

            # if we are at the end of the values then there no more children and we see if we consumed all the right bytes
            if values_pos == len(item.values):
                return (
                    state_set_pos == item.start
                )  # note that ".start" mean end because items are reversed

            # get the child we are trying to match (meaning we are looking for completed early items for this node)
            value = item.values[values_pos]

            # if we have a terminal node we can jump forward that many bytes
            if isinstance(value, Terminal):
                item.children[values_pos] = value
                values_pos += 1
                state_set_pos += len(value)
            else:
                break

        # otherwise we need to try all possible next matching items in the current state set
        # so we loop over every item in the current state set looking for a completed match
        for inner_item in reversed_state_sets[state_set_pos]:
            if inner_item.node == value and inner_item.pos == len(inner_item.values):

                # see if we can get a complete parse following this inner item
                if self._compute_children(
                    inner_item.start, item, reversed_state_sets, values_pos + 1
                ):
                    item.children[values_pos] = inner_item
                    return True

        # if we didn't find a child set and this is nullable we can skip this child (since it may not exist if nulled)
        if value.nullable:
            if self._compute_children(
                state_set_pos, item, reversed_state_sets, values_pos + 1
            ):
                item.children[values_pos] = (
                    None  # this child was skipped since it was nullable
                )
                return True

        return False



import os
import typing

import diskcache as dc
import hashlib
import platformdirs
import tiktoken


from ._model import Chat, Instruct
from ._grammarless import GrammarlessEngine, Grammarless

try:
    import openai

    client_class: typing.Optional[typing.Type[openai.OpenAI]] = openai.OpenAI
except ImportError:
    client_class = None


class OpenAIEngine(GrammarlessEngine):
    def __init__(
        self,
        tokenizer,
        max_streaming_tokens,
        timeout,
        compute_log_probs,
        model,
        client_class=client_class,
        **kwargs,
    ):

        if client_class is None:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            )

        self.client = client_class(**kwargs)
        self.model_name = model

        if tokenizer is None:
            tokenizer = tiktoken.encoding_for_model(model)

        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)


class OpenAI(Grammarless):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        api_key=None,
        max_streaming_tokens=1000,
        timeout=0.5,
        compute_log_probs=False,
        engine_class=None,
        **kwargs,
    ):
        """Build a new OpenAI model object that represents a model in a given state.

        This class automatically subclasses itself into the appropriate OpenAIChat, OpenAIInstruct,
        or OpenAICompletion subclass based on the model name.

        Parameters
        ----------
        model : str
            The name of the OpenAI model to use (e.g. gpt-3.5-turbo).
        tokenizer : None or tiktoken.Encoding
            The tokenizer to use for the given model. If set to None we use `tiktoken.encoding_for_model(model)`.
        echo : bool
            If true the final result of creating this model state will be displayed (as HTML in a notebook).
        api_key : None or str
            The OpenAI API key to use for remote requests, passed directly to the `openai.OpenAI` constructor.
        max_streaming_tokens : int
            The maximum number of tokens we allow this model to generate in a single stream. Normally this is set very
            high and we rely either on early stopping on the remote side, or on the grammar terminating causing the
            stream loop to break on the local side. This number needs to be longer than the longest stream you want
            to generate.
        **kwargs :
            All extra keyword arguments are passed directly to the `openai.OpenAI` constructor. Commonly used argument
            names include `base_url` and `organization`
        """

        if client_class is None:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            )

        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is OpenAI:

            # instruct
            # elif "instruct" in model: # All current OpenAI instruct models behave as Completion models.
            #     found_subclass = OpenAIInstruct
                
            found_subclass: typing.Type[OpenAI] = (
                OpenAICompletion
                if model.endswith("-instruct") 
                else OpenAIChat
            )

            # convert to any found subclass
            self.__class__ = found_subclass
            found_subclass.__init__(
                self,
                model,
                tokenizer=tokenizer,
                echo=echo,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                **kwargs,
            )
            return  # we return since we just ran init above and don't need to run again

        # this allows us to use a single constructor for all our subclasses
        if engine_class is None:
            engine_map = {
                OpenAICompletion: OpenAICompletionEngine,
                OpenAIInstruct: OpenAIInstructEngine,
                OpenAIChat: OpenAIChatEngine,
            }
            for k in engine_map:
                if issubclass(self.__class__, k):
                    engine_class = engine_map[k]
                    break

        super().__init__(
            engine_class(
                tokenizer=tokenizer,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                model=model,
                **kwargs,
            ),
            echo=echo,
        )


class OpenAICompletion(OpenAI):
    pass


class OpenAICompletionEngine(OpenAIEngine):
    def _generator(self, prompt, temperature):

        self._reset_shared_data(prompt, temperature)  # update our shared data state

        try:
            generator = self.client.completions.create(
                model=self.model_name,
                prompt=prompt.decode("utf8"),
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1.0,  # TODO: this should be controllable like temp (from the grammar)
                temperature=temperature,
                stream=True,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for part in generator:
            if len(part.choices) > 0:
                chunk = part.choices[0].text or ""
            else:
                chunk = ""
            yield chunk.encode("utf8")


class OpenAIInstruct(OpenAI, Instruct):
    def get_role_start(self, name):
        return ""

    def get_role_end(self, name):
        if name == "instruction":
            return "<|endofprompt|>"
        else:
            raise Exception(
                f"The OpenAIInstruct model does not know about the {name} role type!"
            )


class OpenAIInstructEngine(OpenAIEngine):
    def _generator(self, prompt, temperature):
        # start the new stream
        eop_count = prompt.count(b"<|endofprompt|>")
        if eop_count > 1:
            raise Exception(
                "This model has been given multiple instruct blocks or <|endofprompt|> tokens, but this is not allowed!"
            )
        updated_prompt = prompt + b"<|endofprompt|>" if eop_count == 0 else prompt

        self._reset_shared_data(updated_prompt, temperature)

        try:
            generator = self.client.completions.create(
                model=self.model_name,
                prompt=self._shared_state["data"].decode("utf8"),
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1.0,  # TODO: this should be controllable like temp (from the grammar)
                temperature=temperature,
                stream=True,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        for part in generator:
            if len(part.choices) > 0:
                chunk = part.choices[0].text or ""
            else:
                chunk = ""
            yield chunk.encode("utf8")


class OpenAIChat(OpenAI, Chat):
    pass


class OpenAIChatEngine(OpenAIEngine):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        path = os.path.join(platformdirs.user_cache_dir("guidance"), "openai.tokens")
        self.cache = dc.Cache(path)

    def _hash_prompt(self, prompt):
        return hashlib.sha256(f"{prompt}".encode()).hexdigest()

    def _generator(self, prompt, temperature):

        # find the role tags
        pos = 0
        role_end = b"<|im_end|>"
        messages = []
        found = True
        while found:

            # find the role text blocks
            found = False
            for role_name, start_bytes in (
                ("system", b"<|im_start|>system\n"),
                ("user", b"<|im_start|>user\n"),
                ("assistant", b"<|im_start|>assistant\n"),
            ):
                if prompt[pos:].startswith(start_bytes):
                    pos += len(start_bytes)
                    end_pos = prompt[pos:].find(role_end)
                    if end_pos < 0:
                        assert (
                            role_name == "assistant"
                        ), "Bad chat format! Last role before gen needs to be assistant!"
                        break
                    btext = prompt[pos : pos + end_pos]
                    pos += end_pos + len(role_end)
                    messages.append(
                        {"role": role_name, "content": btext.decode("utf8")}
                    )
                    found = True
                    break

        # Add nice exception if no role tags were used in the prompt.
        # TODO: Move this somewhere more general for all chat models?
        if messages == []:
            raise ValueError(
                f"The OpenAI model {self.model_name} is a Chat-based model and requires role tags in the prompt! \
            Make sure you are using guidance context managers like `with system():`, `with user():` and `with assistant():` \
            to appropriately format your guidance program for this type of model."
            )

        # Update shared data state
        self._reset_shared_data(prompt[:pos], temperature)

        # Use cache only when temperature is 0
        if temperature == 0:
            cache_key = self._hash_prompt(prompt)

            # Check if the result is already in the cache
            if cache_key in self.cache:
                for chunk in self.cache[cache_key]:
                    yield chunk
                return

        # API call and response handling
        try:
            generator = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1.0,  # TODO: this should be controllable like temp (from the grammar)
                temperature=temperature,
                stream=True,
            )

            if temperature == 0:
                cached_results = []

            for part in generator:
                if len(part.choices) > 0:
                    chunk = part.choices[0].delta.content or ""
                else:
                    chunk = ""
                encoded_chunk = chunk.encode("utf8")
                yield encoded_chunk

                if temperature == 0:
                    cached_results.append(encoded_chunk)

            # Cache the results after the generator is exhausted
            if temperature == 0:
                self.cache[cache_key] = cached_results

        except Exception as e:  # TODO: add retry logic
            raise e



from json import dumps as json_dumps
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Union,
    Type,
    TYPE_CHECKING,
)

try:
    import jsonschema
    import pydantic
except ImportError:
    if TYPE_CHECKING:
        raise

from .._guidance import guidance
from ..library import char_range, one_or_more, optional, zero_or_more

from .._grammar import GrammarFunction, select, capture, with_temperature
from ._pydantic import pydantic_to_json_schema


def _to_compact_json(target: Any) -> str:
    # See 'Compact Encoding':
    # https://docs.python.org/3/library/json.html
    # Since this is ultimately about the generated
    # output, we don't need to worry about pretty printing
    # and whitespace
    return json_dumps(target, separators=(",", ":"))


_DEFS_KEYS = ["$defs", "definitions"]


@guidance(stateless=True)
def _gen_json_int(lm):
    pos_nonzero = char_range("1", "9") + zero_or_more(char_range("0", "9"))
    return lm + optional("-") + select(["0", pos_nonzero])


@guidance(stateless=True)
def _gen_json_number(lm):
    mantissa_int = _gen_json_int()
    mantissa_frac = "." + one_or_more(char_range("0", "9"))
    exponent = "e" + select(["", "+", "-"]) + one_or_more(char_range("0", "9"))

    return lm + mantissa_int + optional(mantissa_frac) + optional(exponent)


@guidance(stateless=True)
def _gen_json_string(lm):
    string_chars = select(
        [
            char_range("a", "z"),
            char_range("A", "Z"),
            char_range("0", "9"),
            *[c for c in "-_' ,.!?/[]{}():;"],
            "\\n",
            "\\t",
            "\\\\",
        ],
        recurse=True,
    )
    return lm + '"' + string_chars + '"'


@guidance(stateless=True)
def _gen_json_object(
    lm,
    *,
    properties: Union[Mapping[str, Any], None],
    additional_properties: Union[Mapping[str, Any], None],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    lm += "{"
    if properties:
        lm += _process_properties(properties=properties, definitions=definitions)
    if properties and additional_properties:
        lm += optional(
            ","
            + _process_additional_properties(
                additional_properties=additional_properties, definitions=definitions
            )
        )
    elif additional_properties:
        lm += optional(
            _process_additional_properties(
                additional_properties=additional_properties, definitions=definitions
            )
        )
    lm += "}"
    return lm


@guidance(stateless=True)
def _process_properties(
    lm,
    *,
    properties: Mapping[str, Any],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    properties_added = 0
    for name, property_schema in properties.items():
        lm += '"' + name + '"'

        lm += ":"
        lm += _gen_json(
            json_schema=property_schema,
            definitions=definitions,
        )
        properties_added += 1
        if properties_added < len(properties):
            lm += ","
    return lm


@guidance(stateless=True)
def _process_additional_properties(
    lm,
    *,
    additional_properties: Mapping[str, Any],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    item = (
        _gen_json_string()
        + ":"
        + _gen_json(json_schema=additional_properties, definitions=definitions)
    )
    return lm + zero_or_more(item + ",") + item


@guidance(stateless=True)
def _gen_json_array(
    lm,
    *,
    prefix_items_schema: Optional[Sequence[Mapping[str, Any]]],
    item_schema: Optional[Mapping[str, Any]],
    min_items: int,
    max_items: Optional[int],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    if prefix_items_schema is None:
        prefix_items_schema = []

    if len(prefix_items_schema) < min_items and item_schema is None:
        raise ValueError(
            "No items schema provided, but prefixItems has too few elements "
            f"({len(prefix_items_schema)}) to satisfy minItems ({min_items})"
        )

    if max_items is not None and max_items < min_items:
        raise ValueError(
            f"maxItems ({max_items}) can't be less than minItems ({min_items})"
        )

    required_items = []
    optional_items = []

    # If max_items is None, we can add an infinite tail of items later
    n_to_add = (
        max(len(prefix_items_schema), min_items) if max_items is None else max_items
    )
    for i in range(n_to_add):
        if i < len(prefix_items_schema):
            schema = prefix_items_schema[i]
        elif item_schema is not None:
            schema = item_schema
        else:
            assert i >= min_items
            break

        item = _gen_json(json_schema=schema, definitions=definitions)

        if i < min_items:
            required_items.append(item)
        else:
            optional_items.append(item)

    if max_items is None and item_schema is not None:
        # Add an infinite tail of items
        item = _gen_json(json_schema=item_schema, definitions=definitions)
        optional_items.append(item + zero_or_more("," + item))

    lm += "["

    if required_items:
        first, *rest = required_items
        lm += first
        for item in rest:
            lm += "," + item

    if optional_items:
        # This is a bit subtle and would not be required if not for prefixItems -- the previous
        # must be present before the next one may be added, meaning we have nested optionals:
        # (first optional(,second optional(,third (optional(,...)))))
        first, *rest = optional_items
        tail = ""
        for item in reversed(rest):
            tail = optional("," + item + tail)
        tail = first + tail

        if required_items:
            lm += optional("," + tail)
        else:
            lm += optional(tail)

    lm += "]"
    return lm


@guidance(stateless=True)
def _process_anyOf(
    lm,
    *,
    anyof_list: Sequence[Mapping[str, Any]],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    options = [
        _gen_json(json_schema=item, definitions=definitions) for item in anyof_list
    ]
    return lm + select(options)


@guidance(stateless=True)
def _process_enum(lm, *, options: Sequence[Mapping[str, Any]]):
    # options will come in as python objects, so we need to convert to (compact) JSON
    all_opts = []
    for opt in options:
        all_opts.append(_to_compact_json(opt))
    return lm + select(options=all_opts)


@guidance(stateless=True)
def _gen_json(
    lm,
    json_schema: Mapping[str, Any],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    ANYOF_STRING = "anyOf"
    if ANYOF_STRING in json_schema:
        return lm + _process_anyOf(
            anyof_list=json_schema[ANYOF_STRING], definitions=definitions
        )

    ALLOF_STRING = "allOf"
    if ALLOF_STRING in json_schema:
        allof_list = json_schema[ALLOF_STRING]
        if len(allof_list) != 1:
            raise ValueError("Only support allOf with exactly one item")
        return lm + _gen_json(allof_list[0], definitions)

    REF_STRING = "$ref"
    if REF_STRING in json_schema:
        return lm + _get_definition(
            reference=json_schema[REF_STRING], definitions=definitions
        )

    ENUM_STRING = "enum"
    if ENUM_STRING in json_schema:
        return lm + _process_enum(options=json_schema["enum"])

    TYPE_STRING = "type"
    if TYPE_STRING in json_schema:
        target_type = json_schema["type"]
        if target_type == "null":
            return lm + "null"
        if target_type == "boolean":
            return lm + select(["true", "false"])
        if target_type == "integer":
            return lm + _gen_json_int()
        if target_type == "number":
            return lm + _gen_json_number()
        if target_type == "string":
            return lm + _gen_json_string()
        if target_type == "array":
            return lm + _gen_json_array(
                prefix_items_schema=json_schema.get("prefixItems"),
                item_schema=json_schema.get("items"),
                min_items=json_schema.get("minItems", 0),
                max_items=json_schema.get("maxItems"),
                definitions=definitions,
            )
        if target_type == "object":
            return lm + _gen_json_object(
                properties=json_schema.get("properties"),
                additional_properties=json_schema.get("additionalProperties"),
                definitions=definitions,
            )
        raise ValueError(f"Unsupported type in schema: {target_type}")

    raise ValueError(f"Can't process JSON node: {json_schema}")


@guidance(stateless=True)
def json(
    lm,
    name: Optional[str] = None,
    *,
    schema: Union[
        Mapping[str, Any],
        Type["pydantic.BaseModel"],
        "pydantic.TypeAdapter",
    ],
    temperature: float = 0.0,
):
    """Generate valid JSON according to the supplied JSON schema or `pydantic` model.

    Not all parts of JSON schema (https://json-schema.org/) are supported. Indeed some parts
    (such as bounds on numbers) cannot really be supported in the context of LLM generation.

    Using a JSON schema:
    >>> schema = ''{ "type": "object", "properties": { "a" : {"type": "integer"} } }'
    >>> schema_obj = json.loads(schema)
    >>> lm += json(name="generated_object", schema=schema_obj)
    >>> print(json.loads(lm["generated_object"]))
    { 'a' : 2 }

    Using a `pydantic.BaseModel`:
    >>> class Schema(BaseModel):
    ...     b: bool
    >>> lm += json(name="generated_object", schema=Schema)
    >>> print(json.loads(lm["generated_object"]))
    { 'b' : False }

    Using a `pydantic.TypeAdapter`:
    >>> schema = TypeAdapter(list[int])
    >>> lm += json(name="generated_object", schema=schema)
    >>> print(json.loads(lm["generated_object"]))
    [1, 2, 3]

    Parameters
    ----------

    name : str or None
        If this is not None then the the results of the generation will be saved as a variable on
        the Model object (so you can access the result as `lm["var_name"]`).

    schema : Union[Mapping[str, Any], Type[pydantic.BaseModel], pydantic.TypeAdapter]
        One of:
            - A JSON schema object. This is a JSON schema string which has been passed to `json.loads()`.
            - A subclass of `pydantic.BaseModel`
            - An instance of `pydantic.TypeAdapter`
    """
    if isinstance(schema, Mapping):
        # Raises jsonschema.exceptions.SchemaError or ValueError
        # if schema is not valid
        jsonschema.validators.Draft202012Validator.check_schema(schema)
    else:
        schema = pydantic_to_json_schema(schema)

    definitions: Mapping[str, Callable[[], GrammarFunction]] = {}
    for dk in _DEFS_KEYS:
        if dk in schema:
            assert len(definitions) == 0, "Found duplicate definitions"
            definitions = _build_definitions(schema[dk])

    return lm + capture(
        with_temperature(_gen_json(schema, definitions), temperature=temperature),
        name=name,
    )


def _build_definitions(
    raw_definitions: Mapping[str, Any]
) -> Mapping[str, Callable[[], GrammarFunction]]:
    definitions: Dict[str, Callable[[], GrammarFunction]] = {}

    def build_definition(
        json_schema: Mapping[str, Any]
    ) -> Callable[[], GrammarFunction]:
        @guidance(stateless=True, dedent=False)
        def closure(lm):
            return lm + _gen_json(json_schema=json_schema, definitions=definitions)

        return closure

    definitions = {
        ref: build_definition(schema) for ref, schema in raw_definitions.items()
    }
    return definitions


@guidance(stateless=True)
def _get_definition(
    lm,
    *,
    reference: str,
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    assert definitions is not None
    target_definition = None
    for dk in _DEFS_KEYS:
        ref_start = f"#/{dk}/"
        if reference.startswith(ref_start):
            target_name = reference[len(ref_start) :]
            target_definition = definitions[target_name]

    assert target_definition is not None
    return lm + target_definition()
